"""Resonant Raman intensities"""

import sys
from pathlib import Path

import numpy as np

import ase.units as u
from ase.parallel import paropen, parprint, world
from ase.vibrations import Vibrations
from ase.vibrations.raman import Raman, RamanCalculatorBase


class ResonantRamanCalculator(RamanCalculatorBase, Vibrations):
    """Base class for resonant Raman calculators using finite differences.
    """

    def __init__(self, atoms, ExcitationsCalculator, *args,
                 exkwargs=None, exext='.ex.gz', overlap=False,
                 **kwargs):
        """
        Parameters
        ----------
        atoms: Atoms
            The Atoms object
        ExcitationsCalculator: object
            Calculator for excited states
        exkwargs: dict
            Arguments given to the ExcitationsCalculator object
        exext: string
            Extension for filenames of Excitation lists (results of
            the ExcitationsCalculator).
        overlap : function or False
            Function to calculate overlaps between excitation at
            equilibrium and at a displaced position. Calculators are
            given as first and second argument, respectively.

        Example
        -------

        >>> from ase.calculators.h2morse import (H2Morse,
        ...                                      H2MorseExcitedStatesCalculator)
        >>> from ase.vibrations.resonant_raman import ResonantRamanCalculator
        >>>
        >>> atoms = H2Morse()
        >>> rmc = ResonantRamanCalculator(atoms, H2MorseExcitedStatesCalculator)
        >>> rmc.run()

        This produces all necessary data for further analysis.
        """
        self.exobj = ExcitationsCalculator
        if exkwargs is None:
            exkwargs = {}
        self.exkwargs = exkwargs
        self.overlap = overlap

        super().__init__(atoms, *args, exext=exext, **kwargs)

    def _new_exobj(self):
        # XXXX I have to duplicate this because there are two objects
        # which have exkwargs, why are they not unified?
        return self.exobj(**self.exkwargs)

    def calculate(self, atoms, disp):
        """Call ground and excited state calculation"""
        assert atoms == self.atoms  # XXX action required
        forces = self.atoms.get_forces()

        if self.overlap:
            """Overlap is determined as

            ov_ij = int dr displaced*_i(r) eqilibrium_j(r)
            """
            ov_nn = self.overlap(self.atoms.calc,
                                 self.eq_calculator)
            if world.rank == 0:
                disp.save_ov_nn(ov_nn)

        disp.calculate_and_save_exlist(atoms)
        return {'forces': forces}

    def run(self):
        if self.overlap:
            # XXXX stupid way to make a copy
            self.atoms.get_potential_energy()
            self.eq_calculator = self.atoms.calc
            Path(self.name).mkdir(parents=True, exist_ok=True)
            fname = Path(self.name) / 'tmp.gpw'
            self.eq_calculator.write(fname, 'all')
            self.eq_calculator = self.eq_calculator.__class__(restart=fname)
            try:
                # XXX GPAW specific
                self.eq_calculator.converge_wave_functions()
            except AttributeError:
                pass
        Vibrations.run(self)


class ResonantRaman(Raman):
    """Base Class for resonant Raman intensities using finite differences.
    """

    def __init__(self, atoms, Excitations, *args,
                 observation=None,
                 form='v',         # form of the dipole operator
                 exkwargs=None,      # kwargs to be passed to Excitations
                 exext='.ex.gz',   # extension for Excitation names
                 overlap=False,
                 minoverlap=0.02,
                 minrep=0.8,
                 comm=world,
                 **kwargs):
        """
        Parameters
        ----------
        atoms: ase Atoms object
        Excitations: class
            Type of the excitation list object. The class object is
            initialized as::

                Excitations(atoms.calc)

            or by reading form a file as::

                Excitations('filename', **exkwargs)

            The file is written by calling the method
            Excitations.write('filename').

            Excitations should work like a list of ex obejects, where:
                ex.get_dipole_me(form='v'):
                    gives the velocity form dipole matrix element in
                    units |e| * Angstrom
                ex.energy:
                    is the transition energy in Hartrees
        approximation: string
            Level of approximation used.
        observation: dict
            Polarization settings
        form: string
            Form of the dipole operator, 'v' for velocity form (default)
            and 'r' for length form.
        overlap: bool or function
            Use wavefunction overlaps.
        minoverlap: float ord dict
            Minimal absolute overlap to consider. Defaults to 0.02 to avoid
            numerical garbage.
        minrep: float
            Minimal representation to consider derivative, defaults to 0.8
        """

        if observation is None:
            observation = {'geometry': '-Z(XX)Z'}

        kwargs['exext'] = exext
        Raman.__init__(self, atoms, *args, **kwargs)
        assert self.vibrations.nfree == 2

        self.exobj = Excitations
        if exkwargs is None:
            exkwargs = {}
        self.exkwargs = exkwargs
        self.observation = observation
        self.dipole_form = form

        self.overlap = overlap
        if not isinstance(minoverlap, dict):
            # assume it's a number
            self.minoverlap = {'orbitals': minoverlap,
                               'excitations': minoverlap}
        else:
            self.minoverlap = minoverlap
        self.minrep = minrep

    def read_exobj(self, filename):
        return self.exobj.read(filename, **self.exkwargs)

    def get_absolute_intensities(self, omega, gamma=0.1, delta=0, **kwargs):
        """Absolute Raman intensity or Raman scattering factor

        Parameter
        ---------
        omega: float
           incoming laser energy, unit eV
        gamma: float
           width (imaginary energy), unit eV
        delta: float
           pre-factor for asymmetric anisotropy, default 0

        References
        ----------
        Porezag and Pederson, PRB 54 (1996) 7830-7836 (delta=0)
        Baiardi and Barone, JCTC 11 (2015) 3267-3280 (delta=5)

        Returns
        -------
        raman intensity, unit Ang**4/amu
        """
        alpha2_r, gamma2_r, delta2_r = self._invariants(
            self.electronic_me_Qcc(omega, gamma, **kwargs))
        return 45 * alpha2_r + delta * delta2_r + 7 * gamma2_r

    @property
    def approximation(self):
        return self._approx

    @approximation.setter
    def approximation(self, value):
        self.set_approximation(value)

    def read_excitations(self):
        """Read all finite difference excitations and select matching."""
        if self.overlap:
            return self.read_excitations_overlap()

        disp = self._eq_disp()
        ex0_object = disp.read_exobj()
        eu = ex0_object.energy_to_eV_scale
        matching = frozenset(ex0_object)

        def append(lst, disp, matching):
            exo = disp.read_exobj()
            lst.append(exo)
            matching = matching.intersection(exo)
            return matching

        exm_object_list = []
        exp_object_list = []
        for a, i in zip(self.myindices, self.myxyz):
            mdisp = self._disp(a, i, -1)
            pdisp = self._disp(a, i, 1)
            matching = append(exm_object_list,
                              mdisp, matching)
            matching = append(exp_object_list,
                              pdisp, matching)

        def select(exl, matching):
            mlst = [ex for ex in exl if ex in matching]
            assert len(mlst) == len(matching)
            return mlst

        ex0 = select(ex0_object, matching)
        exm = []
        exp = []
        r = 0
        for a, i in zip(self.myindices, self.myxyz):
            exm.append(select(exm_object_list[r], matching))
            exp.append(select(exp_object_list[r], matching))
            r += 1

        self.ex0E_p = np.array([ex.energy * eu for ex in ex0])
        self.ex0m_pc = (np.array(
            [ex.get_dipole_me(form=self.dipole_form) for ex in ex0]) *
            u.Bohr)
        exmE_rp = []
        expE_rp = []
        exF_rp = []
        exmm_rpc = []
        expm_rpc = []
        r = 0
        for a, i in zip(self.myindices, self.myxyz):
            exmE_rp.append([em.energy for em in exm[r]])
            expE_rp.append([ep.energy for ep in exp[r]])
            exF_rp.append(
                [(em.energy - ep.energy)
                 for ep, em in zip(exp[r], exm[r])])
            exmm_rpc.append(
                [ex.get_dipole_me(form=self.dipole_form)
                 for ex in exm[r]])
            expm_rpc.append(
                [ex.get_dipole_me(form=self.dipole_form)
                 for ex in exp[r]])
            r += 1
        # indicees: r=coordinate, p=excitation
        # energies in eV
        self.exmE_rp = np.array(exmE_rp) * eu
        self.expE_rp = np.array(expE_rp) * eu
        # forces in eV / Angstrom
        self.exF_rp = np.array(exF_rp) * eu / 2 / self.delta
        # matrix elements in e * Angstrom
        self.exmm_rpc = np.array(exmm_rpc) * u.Bohr
        self.expm_rpc = np.array(expm_rpc) * u.Bohr

    def read_excitations_overlap(self):
        """Read all finite difference excitations and wf overlaps.

        We assume that the wave function overlaps are determined as

        ov_ij = int dr displaced*_i(r) eqilibrium_j(r)
        """
        ex0 = self._eq_disp().read_exobj()
        eu = ex0.energy_to_eV_scale
        rep0_p = np.ones((len(ex0)), dtype=float)

        def load(disp, rep0_p):
            ex_p = disp.read_exobj()
            ov_nn = disp.load_ov_nn()
            # remove numerical garbage
            ov_nn = np.where(np.abs(ov_nn) > self.minoverlap['orbitals'],
                             ov_nn, 0)
            ov_pp = ex_p.overlap(ov_nn, ex0)
            ov_pp = np.where(np.abs(ov_pp) > self.minoverlap['excitations'],
                             ov_pp, 0)
            rep0_p *= (ov_pp.real**2 + ov_pp.imag**2).sum(axis=0)
            return ex_p, ov_pp

        def rotate(ex_p, ov_pp):
            e_p = np.array([ex.energy for ex in ex_p])
            m_pc = np.array(
                [ex.get_dipole_me(form=self.dipole_form) for ex in ex_p])
            r_pp = ov_pp.T
            return ((r_pp.real**2 + r_pp.imag**2).dot(e_p),
                    r_pp.dot(m_pc))

        exmE_rp = []
        expE_rp = []
        exF_rp = []
        exmm_rpc = []
        expm_rpc = []
        exdmdr_rpc = []
        for a, i in zip(self.myindices, self.myxyz):
            mdisp = self._disp(a, i, -1)
            pdisp = self._disp(a, i, 1)
            ex, ov = load(mdisp, rep0_p)
            exmE_p, exmm_pc = rotate(ex, ov)
            ex, ov = load(pdisp, rep0_p)
            expE_p, expm_pc = rotate(ex, ov)
            exmE_rp.append(exmE_p)
            expE_rp.append(expE_p)
            exF_rp.append(exmE_p - expE_p)
            exmm_rpc.append(exmm_pc)
            expm_rpc.append(expm_pc)
            exdmdr_rpc.append(expm_pc - exmm_pc)

        # select only excitations that are sufficiently represented
        self.comm.product(rep0_p)
        select = np.where(rep0_p > self.minrep)[0]

        self.ex0E_p = np.array([ex.energy * eu for ex in ex0])[select]
        self.ex0m_pc = (np.array(
            [ex.get_dipole_me(form=self.dipole_form)
             for ex in ex0])[select] * u.Bohr)

        if len(self.myr):
            # indicees: r=coordinate, p=excitation
            # energies in eV
            self.exmE_rp = np.array(exmE_rp)[:, select] * eu
            self.expE_rp = np.array(expE_rp)[:, select] * eu
            # forces in eV / Angstrom
            self.exF_rp = np.array(exF_rp)[:, select] * eu / 2 / self.delta
            # matrix elements in e * Angstrom
            self.exmm_rpc = np.array(exmm_rpc)[:, select, :] * u.Bohr
            self.expm_rpc = np.array(expm_rpc)[:, select, :] * u.Bohr
            # matrix element derivatives in e
            self.exdmdr_rpc = (np.array(exdmdr_rpc)[:, select, :] *
                               u.Bohr / 2 / self.delta)
        else:
            # did not read
            self.exmE_rp = self.expE_rp = self.exF_rp = np.empty(0)
            self.exmm_rpc = self.expm_rpc = self.exdmdr_rpc = np.empty(0)

    def read(self, *args, **kwargs):
        """Read data from a pre-performed calculation."""
        self.vibrations.read(*args, **kwargs)
        self.init_parallel_read()
        if not hasattr(self, 'ex0E_p'):
            if self.overlap:
                self.read_excitations_overlap()
            else:
                self.read_excitations()

        self._already_read = True

    def get_cross_sections(self, omega, gamma):
        """Returns Raman cross sections for each vibration."""
        I_v = self.intensity(omega, gamma)
        pre = 1. / 16 / np.pi**2 / u._eps0**2 / u._c**4
        # frequency of scattered light
        omS_v = omega - self.om_Q
        return pre * omega * omS_v**3 * I_v

    def get_spectrum(self, omega, gamma=0.1,
                     start=None, end=None, npts=None, width=20,
                     type='Gaussian',
                     intensity_unit='????', normalize=False):
        """Get resonant Raman spectrum.

        The method returns wavenumbers in cm^-1 with corresponding
        Raman cross section.
        Start and end point, and width of the Gaussian/Lorentzian should
        be given in cm^-1.
        """

        self.type = type.lower()
        assert self.type in ['gaussian', 'lorentzian']

        frequencies = self.get_energies().real / u.invcm
        intensities = self.get_cross_sections(omega, gamma)
        if width is None:
            return [frequencies, intensities]

        if start is None:
            start = min(self.om_Q) / u.invcm - 3 * width
        if end is None:
            end = max(self.om_Q) / u.invcm + 3 * width

        if not npts:
            npts = int((end - start) / width * 10 + 1)

        prefactor = 1
        if self.type == 'lorentzian':
            intensities = intensities * width * np.pi / 2.
            if normalize:
                prefactor = 2. / width / np.pi
        else:
            sigma = width / 2. / np.sqrt(2. * np.log(2.))
            if normalize:
                prefactor = 1. / sigma / np.sqrt(2 * np.pi)
        # Make array with spectrum data
        spectrum = np.empty(npts)
        energies = np.linspace(start, end, npts)
        for i, energy in enumerate(energies):
            energies[i] = energy
            if self.type == 'lorentzian':
                spectrum[i] = (intensities * 0.5 * width / np.pi /
                               ((frequencies - energy)**2 +
                                0.25 * width**2)).sum()
            else:
                spectrum[i] = (intensities *
                               np.exp(-(frequencies - energy)**2 /
                                      2. / sigma**2)).sum()
        return [energies, prefactor * spectrum]

    def write_spectrum(self, omega, gamma,
                       out='resonant-raman-spectra.dat',
                       start=200, end=4000,
                       npts=None, width=10,
                       type='Gaussian'):
        """Write out spectrum to file.

        Start and end
        point, and width of the Gaussian/Lorentzian should be given
        in cm^-1."""
        energies, spectrum = self.get_spectrum(omega, gamma,
                                               start, end, npts, width,
                                               type)

        # Write out spectrum in file. First column is absolute intensities.
        outdata = np.empty([len(energies), 3])
        outdata.T[0] = energies
        outdata.T[1] = spectrum

        with paropen(out, 'w') as fd:
            fd.write('# Resonant Raman spectrum\n')
            if hasattr(self, '_approx'):
                fd.write(f'# approximation: {self._approx}\n')
            for key in self.observation:
                fd.write(f'# {key}: {self.observation[key]}\n')
            fd.write('# omega={:g} eV, gamma={:g} eV\n'
                     .format(omega, gamma))
            if width is not None:
                fd.write('# %s folded, width=%g cm^-1\n'
                         % (type.title(), width))
            fd.write('# [cm^-1]  [a.u.]\n')

            for row in outdata:
                fd.write('%.3f  %15.5g\n' %
                         (row[0], row[1]))

    def summary(self, omega, gamma=0.1,
                method='standard', direction='central',
                log=sys.stdout):
        """Print summary for given omega [eV]"""
        self.read(method, direction)
        hnu = self.get_energies()
        intensities = self.get_absolute_intensities(omega, gamma)
        te = int(np.log10(intensities.max())) - 2
        scale = 10**(-te)
        if not te:
            ts = ''
        elif te > -2 and te < 3:
            ts = str(10**te)
        else:
            ts = f'10^{te}'

        if isinstance(log, str):
            log = paropen(log, 'a')

        parprint('-------------------------------------', file=log)
        parprint(' excitation at ' + str(omega) + ' eV', file=log)
        parprint(' gamma ' + str(gamma) + ' eV', file=log)
        parprint(' method:', self.vibrations.method, file=log)
        parprint(' approximation:', self.approximation, file=log)
        parprint(' Mode    Frequency        Intensity', file=log)
        parprint(f'  #    meV     cm^-1      [{ts}A^4/amu]', file=log)
        parprint('-------------------------------------', file=log)
        for n, e in enumerate(hnu):
            if e.imag != 0:
                c = 'i'
                e = e.imag
            else:
                c = ' '
                e = e.real
            parprint('%3d %6.1f%s  %7.1f%s  %9.2f' %
                     (n, 1000 * e, c, e / u.invcm, c, intensities[n] * scale),
                     file=log)
        parprint('-------------------------------------', file=log)
        parprint('Zero-point energy: %.3f eV' %
                 self.vibrations.get_zero_point_energy(),
                 file=log)


class LrResonantRaman(ResonantRaman):
    """Resonant Raman for linear response

    Quick and dirty approach to enable loading of LrTDDFT calculations
    """

    def read_excitations(self):
        eq_disp = self._eq_disp()
        ex0_object = eq_disp.read_exobj()
        eu = ex0_object.energy_to_eV_scale
        matching = frozenset(ex0_object.kss)

        def append(lst, disp, matching):
            exo = disp.read_exobj()
            lst.append(exo)
            matching = matching.intersection(exo.kss)
            return matching

        exm_object_list = []
        exp_object_list = []
        for a in self.indices:
            for i in 'xyz':
                disp1 = self._disp(a, i, -1)
                disp2 = self._disp(a, i, 1)

                matching = append(exm_object_list,
                                  disp1,
                                  matching)
                matching = append(exp_object_list,
                                  disp2,
                                  matching)

        def select(exl, matching):
            exl.diagonalize(**self.exkwargs)
            mlist = list(exl)
#            mlst = [ex for ex in exl if ex in matching]
#            assert(len(mlst) == len(matching))
            return mlist
        ex0 = select(ex0_object, matching)
        exm = []
        exp = []
        r = 0
        for a in self.indices:
            for i in 'xyz':
                exm.append(select(exm_object_list[r], matching))
                exp.append(select(exp_object_list[r], matching))
                r += 1

        self.ex0E_p = np.array([ex.energy * eu for ex in ex0])
#        self.exmE_p = np.array([ex.energy * eu for ex in exm])
#        self.expE_p = np.array([ex.energy * eu for ex in exp])
        self.ex0m_pc = (np.array(
            [ex.get_dipole_me(form=self.dipole_form) for ex in ex0]) *
            u.Bohr)
        self.exF_rp = []
        exmE_rp = []
        expE_rp = []
        exmm_rpc = []
        expm_rpc = []
        r = 0
        for a in self.indices:
            for i in 'xyz':
                exmE_rp.append([em.energy for em in exm[r]])
                expE_rp.append([ep.energy for ep in exp[r]])
                self.exF_rp.append(
                    [(em.energy - ep.energy)
                     for ep, em in zip(exp[r], exm[r])])
                exmm_rpc.append(
                    [ex.get_dipole_me(form=self.dipole_form) for ex in exm[r]])
                expm_rpc.append(
                    [ex.get_dipole_me(form=self.dipole_form) for ex in exp[r]])
                r += 1
        self.exmE_rp = np.array(exmE_rp) * eu
        self.expE_rp = np.array(expE_rp) * eu
        self.exF_rp = np.array(self.exF_rp) * eu / 2 / self.delta
        self.exmm_rpc = np.array(exmm_rpc) * u.Bohr
        self.expm_rpc = np.array(expm_rpc) * u.Bohr



import json

import numpy as np
import pytest

from ase.build import bulk
from ase.calculators.morse import MorsePotential
from ase.constraints import FixBondLength
from ase.geometry.geometry import find_mic, get_distances
from ase.mep import NEB, NEBTools
from ase.mep.neb import NEBOptimizer
from ase.optimize import BFGS, ODE12r
from ase.optimize.precon import Exp
from ase.utils.forcecurve import fit_images

pytestmark = pytest.mark.optimize


def calc():
    return MorsePotential(A=4.0, epsilon=1.0, r0=2.55)


@pytest.fixture(scope='module')
def _setup_images_global():
    N_intermediate = 3
    N_cell = 2
    initial = bulk('Cu', cubic=True)
    initial *= N_cell

    # place vacancy near centre of cell
    D, D_len = get_distances(np.diag(initial.cell) / 2,
                             initial.positions,
                             initial.cell, initial.pbc)
    vac_index = D_len.argmin()
    vac_pos = initial.positions[vac_index]
    del initial[vac_index]

    # identify two opposing nearest neighbours of the vacancy
    D, D_len = get_distances(vac_pos,
                             initial.positions,
                             initial.cell, initial.pbc)
    D = D[0, :]
    D_len = D_len[0, :]

    nn_mask = np.abs(D_len - D_len.min()) < 1e-8
    i1 = nn_mask.nonzero()[0][0]
    i2 = ((D + D[i1])**2).sum(axis=1).argmin()

    print(f'vac_index={vac_index} i1={i1} i2={i2} '
          f'distance={initial.get_distance(i1, i2, mic=True)}')

    final = initial.copy()
    final.positions[i1] = vac_pos

    initial.calc = calc()
    final.calc = calc()

    qn = ODE12r(initial)
    qn.run(fmax=1e-3)
    qn = ODE12r(final)
    qn.run(fmax=1e-3)

    images = [initial]
    for image in range(N_intermediate):
        image = initial.copy()
        image.calc = calc()
        images.append(image)
    images.append(final)

    neb = NEB(images)
    neb.interpolate()

    return neb.images, i1, i2


@pytest.fixture()
def setup_images(_setup_images_global):
    images, i1, i2 = _setup_images_global
    new_images = [img.copy() for img in images]
    for img in new_images:
        img.calc = calc()
    return new_images, i1, i2


@pytest.fixture(scope='module')
def _ref_vacancy_global(_setup_images_global):
    # use distance from moving atom to one of its neighbours as reaction coord
    # relax intermediate image to the saddle point using a bondlength constraint
    images, i1, i2 = _setup_images_global
    initial, saddle, final = (images[0].copy(),
                              images[2].copy(),
                              images[4].copy())
    initial.calc = calc()
    saddle.calc = calc()
    final.calc = calc()
    saddle.set_constraint(FixBondLength(i1, i2))
    opt = ODE12r(saddle)
    opt.run(fmax=1e-2)
    nebtools = NEBTools([initial, saddle, final])
    Ef_ref, dE_ref = nebtools.get_barrier(fit=False)
    print('REF:', Ef_ref, dE_ref)
    return Ef_ref, dE_ref, saddle


@pytest.fixture()
def ref_vacancy(_ref_vacancy_global):
    Ef_ref, dE_ref, saddle = _ref_vacancy_global
    return Ef_ref, dE_ref, saddle.copy()


@pytest.mark.slow()
@pytest.mark.filterwarnings('ignore:estimate_mu')
@pytest.mark.parametrize('method, optimizer, precon, optmethod',
                         [('aseneb', BFGS, None, None),
                          ('improvedtangent', BFGS, None, None),
                          ('spline', NEBOptimizer, None, 'ODE'),
                          ('string', NEBOptimizer, 'Exp', 'ODE')])
def test_neb_methods(testdir, method, optimizer, precon,
                     optmethod, ref_vacancy, setup_images):
    # unpack the reference result
    Ef_ref, dE_ref, saddle_ref = ref_vacancy

    # now relax the MEP for comparison
    images, _, _ = setup_images

    fmax_history = []

    def save_fmax_history(mep):
        fmax_history.append(mep.get_residual())

    k = 0.1
    if precon == 'Exp':
        k = 0.01
    mep = NEB(images, k=k, method=method, precon=precon)

    if optmethod is not None:
        opt = optimizer(mep, method=optmethod)
    else:
        opt = optimizer(mep)
    opt.attach(save_fmax_history, 1, mep)
    opt.run(fmax=1e-2)

    nebtools = NEBTools(images)
    Ef, dE = nebtools.get_barrier(fit=False)
    print(f'{method},{optimizer.__name__},{precon} '
          f'=> Ef = {Ef:.3f}, dE = {dE:.3f}')

    forcefit = fit_images(images)

    with open(f'MEP_{method}_{optimizer.__name__}_{optmethod}'
              f'_{precon}.json', 'w') as fd:
        json.dump({'fmax_history': fmax_history,
                   'method': method,
                   'optmethod': optmethod,
                   'precon': precon,
                   'optimizer': optimizer.__name__,
                   'path': forcefit.path,
                   'energies': forcefit.energies.tolist(),
                   'fit_path': forcefit.fit_path.tolist(),
                   'fit_energies': forcefit.fit_energies.tolist(),
                   'lines': np.array(forcefit.lines).tolist(),
                   'Ef': Ef,
                   'dE': dE}, fd)

    centre = 2  # we have 5 images total, so central image has index 2
    vdiff, _ = find_mic(images[centre].positions - saddle_ref.positions,
                        images[centre].cell)
    print(f'Ef error {Ef - Ef_ref} dE error {dE - dE_ref} '
          f'position error at saddle {abs(vdiff).max()}')
    assert abs(Ef - Ef_ref) < 1e-2
    assert abs(dE - dE_ref) < 1e-2
    assert abs(vdiff).max() < 1e-2


@pytest.mark.parametrize('method', ['ODE', 'static'])
@pytest.mark.filterwarnings('ignore:NEBOptimizer did not converge')
def test_neb_optimizers(setup_images, method):
    images, _, _ = setup_images
    mep = NEB(images, method='spline', precon='Exp')
    mep.get_forces()  # needed so residuals are available
    R0 = mep.get_residual()
    opt = NEBOptimizer(mep, method=method)
    opt.run(steps=2)  # take two steps
    R1 = mep.get_residual()
    # check residual has got smaller
    assert R1 < R0


def test_precon_initialisation(setup_images):
    images, _, _ = setup_images
    mep = NEB(images, method='spline', precon='Exp')
    mep.get_forces()
    assert len(mep.precon) == len(mep.images)
    assert mep.precon[0].mu == mep.precon[1].mu


def test_single_precon_initialisation(setup_images):
    images, _, _ = setup_images
    precon = Exp()
    mep = NEB(images, method='spline', precon=precon)
    mep.get_forces()
    assert len(mep.precon) == len(mep.images)
    assert mep.precon[0].mu == mep.precon[1].mu


def test_list_precon_initialisation(setup_images):
    images, _, _ = setup_images

    precon = Exp()
    mep = NEB(images, method='spline', precon=precon)
    mep.get_forces()

    # the tested scenario is saving computed precon
    # for restarting of a calculation

    # saving as PreconImages object
    mep_restart = NEB(images, method='spline', precon=mep.precon)
    mep_restart.get_forces()
    assert len(mep_restart.precon) == len(mep_restart.images)
    assert mep_restart.precon[0].mu == mep_restart.precon[1].mu
    # saving as a list of precon objects
    mep_restart = NEB(images, method='spline', precon=mep.precon.precon)
    mep_restart.get_forces()
    assert len(mep_restart.precon) == len(mep_restart.images)
    assert mep_restart.precon[0].mu == mep_restart.precon[1].mu


def test_precon_assembly(setup_images):
    images, _, _ = setup_images
    neb = NEB(images, method='spline', precon='Exp')
    neb.get_forces()  # trigger precon assembly

    # check precon for each image is symmetric positive definite
    for image, precon in zip(neb.images, neb.precon):
        assert isinstance(precon, Exp)
        P = precon.asarray()
        N = 3 * len(image)
        assert P.shape == (N, N)
        assert np.abs(P - P.T).max() < 1e-6
        assert np.all(np.linalg.eigvalsh(P)) > 0


def test_spline_fit(setup_images):
    images, _, _ = setup_images
    neb = NEB(images)
    fit = neb.spline_fit()

    # check spline points are equally spaced
    assert np.allclose(fit.s, np.linspace(0, 1, len(images)))

    # check spline matches target at fit points
    assert np.allclose(fit.x(fit.s), fit.x_data)

    # ensure derivative is smooth across central fit point
    eps = 1e-4
    assert np.allclose(fit.dx_ds(fit.s[2] + eps), fit.dx_ds(fit.s[2] + eps))


def test_integrate_forces(setup_images):
    images, _, _ = setup_images
    forcefit = fit_images(images)

    neb = NEB(images)
    spline_points = 1000  # it is the default value
    _s, E, _F = neb.integrate_forces(spline_points=spline_points)
    # check the difference between initial and final images
    np.testing.assert_allclose(E[0] - E[-1],
                               forcefit.energies[0] - forcefit.energies[-1],
                               atol=1.0e-10)
    # assert the maximum Energy value is in the middle
    assert np.argmax(E) == spline_points // 2 - 1
    # check the maximum values (barrier value)
    # tolerance value is rather high since the images are not relaxed
    np.testing.assert_allclose(E.max(),
                               forcefit.energies.max(), rtol=2.5e-2)



import re
from collections import OrderedDict

import numpy as np

from ase import Atoms
from ase.io import write


def test_xsd_bond():
    # This writes xsd example with bond connectivity information, and checks
    # bond formats.

    # Example molecule
    atoms = Atoms('CH4', [[1.08288111e-09, 1.74602682e-09, -1.54703448e-09],
                          [-6.78446715e-01, 8.73516584e-01, -8.63073811e-02],
                          [-4.09602527e-01, -8.46016530e-01, -5.89280858e-01],
                          [8.52016070e-02, -2.98243876e-01, 1.06515792e+00],
                          [1.00284763e+00, 2.70743821e-01, -3.89569679e-01]])
    # Carbon(index 0), is connected to other hydrogen atoms (index 1-4)
    connectivitymatrix = np.array([[0, 1, 1, 1, 1],
                                   [1, 0, 0, 0, 0],
                                   [1, 0, 0, 0, 0],
                                   [1, 0, 0, 0, 0],
                                   [1, 0, 0, 0, 0]])
    write('xsd_test_CH4.xsd', atoms, connectivity=connectivitymatrix)

    # Read and see if the atom information and bond information matches.
    AtomIdsToBondIds = OrderedDict()
    BondIdsToConnectedAtomIds = OrderedDict()
    with open('xsd_test_CH4.xsd') as fd:
        for i, line in enumerate(fd):
            if '<Atom3d ' in line:
                AtomId = int(re.search(r'ID="(.*?)"', line).group(1))
                ConnectedBondIds = [
                    int(a) for a in re.search(
                        r'Connections="(.*?)"',
                        line).group(1).split(',')]
                AtomIdsToBondIds[AtomId] = ConnectedBondIds
            elif '<Bond ' in line:
                BondId = int(re.search(r'ID="(.*?)"', line).group(1))
                ConnectedAtomIds = [
                    int(a) for a in re.search(
                        r'Connects="(.*?)"',
                        line).group(1).split(',')]
                BondIdsToConnectedAtomIds[BondId] = ConnectedAtomIds
    # check if atom ids have been correctly assigned for each bond
    for AtomId in AtomIdsToBondIds:
        for BondId in AtomIdsToBondIds[AtomId]:
            assert AtomId in BondIdsToConnectedAtomIds[BondId]

    # make connectivity graph and see if it matches with input.
    AtomIds = list(AtomIdsToBondIds.keys())
    Newconnectivitymatrix = np.zeros((5, 5))
    for AtomId in AtomIdsToBondIds:
        for BondId in AtomIdsToBondIds[AtomId]:
            OtherAtomId = [
                a for a in BondIdsToConnectedAtomIds[BondId] if a != AtomId]
            i = AtomIds.index(AtomId)
            j = AtomIds.index(OtherAtomId[0])
            Newconnectivitymatrix[i, j] = 1
    for i in range(4):
        for j in range(4):
            assert connectivitymatrix[i, j] == Newconnectivitymatrix[i, j]



"""Utility for checking Python module imports triggered by any code snippet.

This module was developed to monitor the import footprint of the ase CLI
command: The CLI command can become unnecessarily slow and unresponsive
if too many modules are imported even before the CLI is launched or
it is known what modules will be actually needed.
See https://gitlab.com/ase/ase/-/issues/1124 for more discussion.

The utility here is general, so it can be used for checking and
monitoring other code snippets too.
"""
import json
import os
import re
import sys
from pprint import pprint
from subprocess import run
from typing import List, Optional, Set


def exec_and_check_modules(expression: str) -> Set[str]:
    """Return modules loaded by the execution of a Python expression.

    Parameters
    ----------
    expression
        Python expression

    Returns
    -------
    Set of module names.
    """
    # Take null outside command to avoid
    # `import os` before expression
    null = os.devnull
    command = ("import sys;"
               f" stdout = sys.stdout; sys.stdout = open({null!r}, 'w');"
               f" {expression};"
               " sys.stdout = stdout;"
               " modules = list(sys.modules);"
               " import json; print(json.dumps(modules))")
    proc = run([sys.executable, '-c', command],
               capture_output=True, universal_newlines=True,
               check=True)
    return set(json.loads(proc.stdout))


def check_imports(expression: str, *,
                  forbidden_modules: List[str] = [],
                  max_module_count: Optional[int] = None,
                  max_nonstdlib_module_count: Optional[int] = None,
                  do_print: bool = False) -> None:
    """Check modules imported by the execution of a Python expression.

    Parameters
    ----------
    expression
        Python expression
    forbidden_modules
        Throws an error if any module in this list was loaded.
    max_module_count
        Throws an error if the number of modules exceeds this value.
    max_nonstdlib_module_count
        Throws an error if the number of non-stdlib modules exceeds this value.
    do_print:
        Print loaded modules if set.
    """
    modules = exec_and_check_modules(expression)

    if do_print:
        print('all modules:')
        pprint(sorted(modules))

    for module_pattern in forbidden_modules:
        r = re.compile(module_pattern)
        for module in modules:
            assert not r.fullmatch(module), \
                f'{module} was imported'

    if max_nonstdlib_module_count is not None:
        assert sys.version_info >= (3, 10), 'Python 3.10+ required'

        nonstdlib_modules = []
        for module in modules:
            if (
                module.split('.')[0]
                in sys.stdlib_module_names  # type: ignore[attr-defined]
            ):
                continue
            nonstdlib_modules.append(module)

        if do_print:
            print('nonstdlib modules:')
            pprint(sorted(nonstdlib_modules))

        module_count = len(nonstdlib_modules)
        assert module_count <= max_nonstdlib_module_count, (
            'too many nonstdlib modules loaded:'
            f' {module_count}/{max_nonstdlib_module_count}'
        )

    if max_module_count is not None:
        module_count = len(modules)
        assert module_count <= max_module_count, \
            f'too many modules loaded: {module_count}/{max_module_count}'


if __name__ == '__main__':
    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument('expression')
    parser.add_argument('--forbidden_modules', nargs='+', default=[])
    parser.add_argument('--max_module_count', type=int, default=None)
    parser.add_argument('--max_nonstdlib_module_count', type=int, default=None)
    parser.add_argument('--do_print', action='store_true')
    args = parser.parse_args()

    check_imports(**vars(args))



"""Tests for Octopus outputs."""
from typing import Any, Dict

import numpy as np
import pytest

from ase.io.octopus.output import read_eigenvalues_file, read_static_info
from ase.units import Bohr, Debye, Hartree


@pytest.fixture(name='info_iron')
def fixture_info_iron(datadir) -> Dict[str, Any]:
    """`info` of 'periodic_systems/25-Fe_polarized.01-gs'"""
    file = datadir / 'octopus/periodic_systems_25-Fe_polarized.01-gs_info'
    with file.open(encoding='utf-8') as fd:
        return read_static_info(fd)


def test_eigenvalues(datadir):
    """Test if the eigenvalues are parsed correctly from `eigenvalues`."""
    file = (
        datadir
        / 'octopus/periodic_systems_03-sodium_chain.02-unocc_eigenvalues'
    )
    with file.open(encoding='utf-8') as fd:
        kptsarr, eigsarr, occsarr = read_eigenvalues_file(fd)
    kptsarr_ref = [
        [0.0, 0.0, 0.0],
        [0.1, 0.0, 0.0],
        [0.2, 0.0, 0.0],
        [0.3, 0.0, 0.0],
        [0.4, 0.0, 0.0],
        [0.5, 0.0, 0.0],
    ]
    eigsarr_ref = [
        [[-3.647610, -1.236399]],
        [[-3.551626, -1.144852]],
        [[-3.264556, -0.870333]],
        [[-2.789110, -0.413463]],
        [[-2.130396, -0.280652]],
        [[-1.367358, -1.210151]],
    ]
    occsarr_ref = [
        [[2.0, 0.0]],
        [[2.0, 0.0]],
        [[2.0, 0.0]],
        [[0.0, 0.0]],
        [[0.0, 0.0]],
        [[0.0, 0.0]],
    ]
    np.testing.assert_allclose(kptsarr, kptsarr_ref)
    np.testing.assert_allclose(eigsarr, eigsarr_ref)
    np.testing.assert_allclose(occsarr, occsarr_ref)


def test_fermi_level(info_iron: Dict[str, Any]):
    """Test if the Fermi level is parsed correctly."""
    efermi_ref = 0.153766 * Hartree
    np.testing.assert_allclose(info_iron['fermi_level'], efermi_ref)


def test_dipole_moment(datadir):
    """Test if the dipole moment is parsed correctly."""
    file = datadir / 'octopus/linear_response_02-h2o_pol_lr.01_h2o_gs_info'
    with file.open(encoding='utf-8') as fd:
        results = read_static_info(fd)
    dipole_ref = np.array((7.45151E-16, 9.30594E-01, 3.24621E-15)) * Debye
    np.testing.assert_allclose(results['dipole'], dipole_ref)


def test_magnetic_moment(info_iron: Dict[str, Any]):
    """Test if the magnetic moment is parsed correctly."""
    magmom_ref = 7.409638
    magmoms_ref = [3.385730, 3.385730]
    np.testing.assert_allclose(info_iron['magmom'], magmom_ref)
    np.testing.assert_allclose(info_iron['magmoms'], magmoms_ref)


def test_stress(datadir):
    """Test if the stress tensor is parsed correctly."""
    file = datadir / 'octopus/periodic_systems_30-stress.05-output_scf_info'
    with file.open(encoding='utf-8') as fd:
        results = read_static_info(fd)
    stress_ref = (
        (-5.100825936E-04, -9.121282047E-16, +4.495864617E-16),
        (-9.121277710E-16, -5.100825936E-04, -1.086184124E-15),
        (+4.496421897E-16, -1.086295146E-15, -5.100825937E-04),
    )
    stress_ref = np.array(stress_ref) * Hartree / Bohr**3
    np.testing.assert_allclose(results['stress'], stress_ref)


def test_kpoints(info_iron: Dict[str, Any]):
    """Test if the kpoints are parsed correctly."""
    nkpts_ref = 4
    kpoint_weights_ref = [0.25, 0.25, 0.25, 0.25]
    ibz_kpoints_ref = [
        [0.0000, 0.0000, 0.0000],
        [0.5000, 0.0000, 0.0000],
        [0.0000, 0.5000, 0.0000],
        [0.5000, 0.5000, 0.0000],
    ]
    assert info_iron['nkpts'] == nkpts_ref
    np.testing.assert_allclose(info_iron['kpoint_weights'], kpoint_weights_ref)
    np.testing.assert_allclose(info_iron['ibz_kpoints'], ibz_kpoints_ref)



import os

import pytest

from ase import Atoms
from ase.calculators.vasp import Vasp


@pytest.fixture()
def atoms_co():
    """Simple atoms object for testing with a single CO molecule"""
    d = 1.14
    atoms = Atoms('CO', positions=[(0, 0, 0), (0, 0, d)], pbc=True)
    atoms.center(vacuum=5)
    return atoms


@pytest.fixture()
def atoms_2co():
    """Simple atoms object for testing with 2x CO molecules"""
    d = 1.14
    atoms = Atoms('CO', positions=[(0, 0, 0), (0, 0, d)], pbc=True)
    atoms.extend(Atoms('CO', positions=[(0, 2, 0), (0, 2, d)]))

    atoms.center(vacuum=5.)
    return atoms


@pytest.fixture()
def atoms_nh3():
    """Simple atoms object for testing with a planar NH3 molecule"""
    d1 = 0.5
    d2 = 0.87
    atoms = Atoms(
        'NH3', positions=[
            (0, 0, 0), (0, d1, d2), (0, -d1, d2), (0, 0, -1.)], pbc=True)
    atoms.center(vacuum=5)
    return atoms


@pytest.fixture()
def mock_vasp_calculate(monkeypatch):
    """Fixture which mocks the VASP run method, so a calculation cannot run.
    Acts as a safeguard for tests which want to test VASP,
    but avoid accidentally launching a calculation"""

    def _mock_run(self, command=None, out=None, directory=None):
        assert False, 'Test attempted to launch a calculation'

    # Patch the calculate and run methods, so we're certain
    # calculations aren't accidentally launched
    monkeypatch.setattr('ase.calculators.vasp.Vasp._run', _mock_run)
    return


@pytest.fixture()
def clear_vasp_envvar(monkeypatch):
    """Clear the environment variables which can be used to launch
    a VASP calculation."""
    for envvar in Vasp.env_commands:
        monkeypatch.delenv(envvar, raising=False)
        assert envvar not in os.environ
    return



import numpy as np

import ase.units as units
from ase import Atoms
from ase.calculators.acn import (
    ACN,
    epsilon_c,
    epsilon_me,
    epsilon_n,
    m_me,
    r_cn,
    r_mec,
    sigma_c,
    sigma_me,
    sigma_n,
)
from ase.calculators.qmmm import EIQMMM, LJInteractionsGeneral, SimpleQMMM
from ase.constraints import FixLinearTriatomic
from ase.optimize import BFGS


def test_qmmm_acn(testdir):
    # From https://www.sciencedirect.com/science/article/pii/S0166128099002079
    eref = 4.9 * units.kcal / units.mol
    dref = 3.368
    aref = 79.1

    sigma = np.array([sigma_me, sigma_c, sigma_n])
    epsilon = np.array([epsilon_me, epsilon_c, epsilon_n])
    inter = LJInteractionsGeneral(sigma, epsilon, sigma, epsilon, 3)

    for calc in [ACN(),
                 SimpleQMMM([0, 1, 2], ACN(), ACN(), ACN()),
                 SimpleQMMM([0, 1, 2], ACN(), ACN(), ACN(), vacuum=3.0),
                 EIQMMM([0, 1, 2], ACN(), ACN(), inter),
                 EIQMMM([0, 1, 2], ACN(), ACN(), inter, vacuum=3.0),
                 EIQMMM([3, 4, 5], ACN(), ACN(), inter, vacuum=3.0)]:
        dimer = Atoms('CCNCCN',
                      [(-r_mec, 0, 0),
                       (0, 0, 0),
                       (r_cn, 0, 0),
                       (r_mec, 3.7, 0),
                       (0, 3.7, 0),
                       (-r_cn, 3.7, 0)])

        masses = dimer.get_masses()
        masses[::3] = m_me
        dimer.set_masses(masses)

        dimer.calc = calc

        fixd = FixLinearTriatomic(triples=[(0, 1, 2), (3, 4, 5)])

        dimer.set_constraint(fixd)

        with BFGS(dimer, maxstep=0.04, trajectory=calc.name + '.traj',
                  logfile=calc.name + 'd.log') as opt:
            opt.run(0.001, steps=1000)

        e0 = dimer.get_potential_energy()
        d0 = dimer.get_distance(1, 4)
        a0 = dimer.get_angle(2, 1, 4)
        fmt = '{0:>25}: {1:.3f} {2:.3f} {3:.1f}'
        print(fmt.format(calc.name, -e0, d0, a0))
        assert abs(e0 + eref) < 0.013
        assert abs(d0 - dref) < 0.224
        assert abs(a0 - aref) < 2.9

    print(fmt.format('reference', eref, dref, aref))



'''These tests ensure that the computed PEC curvature matche the actual
geometries using a somewhat agressive angle_limit for each stepsize.'''
import numpy as np
import pytest

from ase import Atoms
from ase.calculators.emt import EMT
from ase.constraints import FixAtoms
from ase.md.contour_exploration import ContourExploration

pair_distance = 2.5


def Al_atom_pair(pair_distance=pair_distance):
    atoms = Atoms('AlAl', positions=[
        [-pair_distance / 2, 0, 0],
        [pair_distance / 2, 0, 0]])
    atoms.center(vacuum=10)
    atoms.calc = EMT()
    return atoms


def test_curvature1(testdir):
    '''This basic test has an atom spinning counter-clockwise around a fixed
    atom. The radius (1/curvature) must therefore be very
    close the pair_distance.'''
    name = 'test_curvature1'

    radius = pair_distance
    atoms = Al_atom_pair(pair_distance)

    atoms.set_constraint(FixAtoms(indices=[0]))
    atoms.set_velocities([[0, 0, 0], [0, 1, 0]])

    with ContourExploration(
        atoms,
        maxstep=1.5,
        parallel_drift=0.0,
        angle_limit=30,
        trajectory=name + '.traj',
        logfile=name + '.log',
    ) as dyn:
        print(f"Target Radius (1/curvature) {radius: .6f} Ang")
        for _ in range(5):
            dyn.run(30)
            print(f'Radius (1/curvature) {1 / dyn.curvature: .6f} Ang')
            assert radius == pytest.approx(1.0 / dyn.curvature, abs=2e-3)


def test_curvature2(testdir):
    '''This test has two atoms spinning counter-clockwise around eachother. the
    The radius (1/curvature) is less obviously pair_distance*sqrt(2)/2.
    This is the simplest multi-body analytic curvature test.'''
    name = 'test_curvature2'

    radius = pair_distance * np.sqrt(2) / 2
    atoms = Al_atom_pair(pair_distance)

    atoms.set_velocities([[0, -1, 0], [0, 1, 0]])

    with ContourExploration(
        atoms,
        maxstep=1.0,
        parallel_drift=0.0,
        angle_limit=30,
        trajectory=name + '.traj',
        logfile=name + '.log',
    ) as dyn:
        print(f"Target Radius (1/curvature) {radius: .6f} Ang")
        for _ in range(5):
            dyn.run(30)
            print(f'Radius (1/curvature) {1 / dyn.curvature: .6f} Ang')
            assert radius == pytest.approx(1.0 / dyn.curvature, abs=2e-3)



"""Module to read and write atoms in xtl file format for the prismatic and
computem software.

See https://prism-em.com/docs-inputs for an example of this format and the
documentation of prismatic.

See https://sourceforge.net/projects/computem/ for the source code of the
computem software.
"""

import numpy as np

from ase.atoms import Atoms, symbols2numbers
from ase.utils import reader

from .utils import verify_cell_for_export, verify_dictionary


@reader
def read_prismatic(fd):
    r"""Import prismatic and computem xyz input file as an Atoms object.

    Reads cell, atom positions, occupancies and Debye Waller factor.
    The occupancy values and the Debye Waller factors are obtained using the
    `get_array` method and the `occupancies` and `debye_waller_factors` keys,
    respectively. The root means square (RMS) values from the
    prismatic/computem xyz file are converted to Debye-Waller factors (B) in Å²
    by:

    .. math::

        B = RMS^2 * 8\pi^2

    """
    # Read comment:
    fd.readline()

    # Read unit cell parameters:
    cellpar = [float(i) for i in fd.readline().split()]

    # Read all data at once
    # Use genfromtxt instead of loadtxt to skip last line
    read_data = np.genfromtxt(fname=fd, skip_footer=1)
    # Convert from RMS to Debye-Waller factor
    RMS = read_data[:, 5]**2 * 8 * np.pi**2

    atoms = Atoms(symbols=read_data[:, 0],
                  positions=read_data[:, 1:4],
                  cell=cellpar,
                  )
    atoms.set_array('occupancies', read_data[:, 4])
    atoms.set_array('debye_waller_factors', RMS)

    return atoms


class XYZPrismaticWriter:
    """ See the docstring of the `write_prismatic` function.
    """

    def __init__(self, atoms, debye_waller_factors=None, comments=None):
        verify_cell_for_export(atoms.get_cell())

        self.atoms = atoms.copy()
        self.atom_types = set(atoms.symbols)
        self.comments = comments

        self.occupancies = self._get_occupancies()
        debye_waller_factors = self._get_debye_waller_factors(
            debye_waller_factors)
        self.RMS = np.sqrt(debye_waller_factors / (8 * np.pi**2))

    def _get_occupancies(self):
        if 'occupancies' in self.atoms.arrays:
            occupancies = self.atoms.get_array('occupancies', copy=False)
        else:
            occupancies = np.ones_like(self.atoms.numbers)

        return occupancies

    def _get_debye_waller_factors(self, DW):
        if np.isscalar(DW):
            if len(self.atom_types) > 1:
                raise ValueError('This cell contains more then one type of '
                                 'atoms and the Debye-Waller factor needs to '
                                 'be provided for each atom using a '
                                 'dictionary.')
            DW = np.ones_like(self.atoms.numbers) * DW
        elif isinstance(DW, dict):
            verify_dictionary(self.atoms, DW, 'DW')
            # Get the arrays of DW from mapping the DW defined by symbol
            DW = {symbols2numbers(k)[0]: v for k, v in DW.items()}
            DW = np.vectorize(DW.get)(self.atoms.numbers)
        else:
            for name in ['DW', 'debye_waller_factors']:
                if name in self.atoms.arrays:
                    DW = self.atoms.get_array(name)

        if DW is None:
            raise ValueError('Missing Debye-Waller factors. It can be '
                             'provided as a dictionary with symbols as key or '
                             'can be set for each atom by using the '
                             '`set_array("debye_waller_factors", values)` of '
                             'the `Atoms` object.')

        return DW

    def _get_file_header(self):
        # 1st line: comment line
        if self.comments is None:
            s = "{} atoms with chemical formula: {}.".format(
                len(self.atoms),
                self.atoms.get_chemical_formula())
        else:
            s = self.comments

        s = s.strip()
        s += " generated by the ase library.\n"
        # 2nd line: lattice parameter
        s += "{} {} {}".format(
            *self.atoms.cell.cellpar()[:3])

        return s

    def write_to_file(self, f):
        data_array = np.vstack((self.atoms.numbers,
                                self.atoms.positions.T,
                                self.occupancies,
                                self.RMS)
                               ).T

        np.savetxt(fname=f,
                   X=data_array,
                   fmt='%.6g',
                   header=self._get_file_header(),
                   newline='\n',
                   footer='-1',
                   comments=''
                   )


def write_prismatic(fd, *args, **kwargs):
    r"""Write xyz input file for the prismatic and computem software. The cell
    needs to be orthorhombric. If the cell contains the `occupancies` and
    `debye_waller_factors` arrays (see the `set_array` method to set them),
    these array will be written to the file.
    If the occupancies is not specified, the default value will be set to 0.

    Parameters:

    atoms: Atoms object

    debye_waller_factors: float or dictionary of float or None (optional).
        If float, set this value to all
        atoms. If dictionary, each atom needs to specified with the symbol as
        key and the corresponding Debye-Waller factor as value.
        If None, the `debye_waller_factors` array of the Atoms object needs to
        be set (see the `set_array` method to set them), otherwise raise a
        ValueError. Since the prismatic/computem software use root means square
        (RMS) displacements, the Debye-Waller factors (B) needs to be provided
        in Å² and these values are converted to RMS displacement by:

        .. math::

            RMS = \sqrt{\frac{B}{8\pi^2}}

        Default is None.

    comment: str (optional)
        Comments to be written in the first line of the file. If not
        provided, write the total number of atoms and the chemical formula.

    """

    writer = XYZPrismaticWriter(*args, **kwargs)
    writer.write_to_file(fd)



import os
import sys
import threading

import numpy as np
import pytest

from ase.calculators.emt import EMT
from ase.calculators.socketio import SocketClient, SocketIOCalculator
from ase.cluster.icosahedron import Icosahedron
from ase.optimize import BFGS

# If multiple test suites are running, we don't want port clashes.
# Thus we generate a port from the pid.
# maxpid is commonly 32768, and max port number is 65536.
# But in case maxpid is much larger for some reason:
pid = os.getpid()
inet_port = (3141 + pid) % 65536
# We could also use a Unix port perhaps, but not yet implemented

# unixsocket = 'grumble'
timeout = 20.0


def getatoms():
    return Icosahedron('Au', 3)


def run_server(launchclient=True, sockettype='unix'):
    atoms = getatoms()

    port = None
    unixsocket = None

    if sockettype == 'unix':
        unixsocket = f'ase_ipi_protocol_bfgs_test_{pid}'
    else:
        assert sockettype == 'inet'
        port = inet_port

    with SocketIOCalculator(log=sys.stdout, port=port,
                            unixsocket=unixsocket,
                            timeout=timeout) as calc:
        if launchclient:
            thread = launch_client_thread(port=port, unixsocket=unixsocket)
        atoms.calc = calc
        with BFGS(atoms) as opt:
            opt.run()

    if launchclient:
        thread.join()

    forces = atoms.get_forces()
    energy = atoms.get_potential_energy()

    atoms.calc = EMT()
    ref_forces = atoms.get_forces()
    ref_energy = atoms.get_potential_energy()

    refatoms = run_normal()
    ref_energy = refatoms.get_potential_energy()
    eerr = abs(energy - ref_energy)
    ferr = np.abs(forces - ref_forces).max()

    perr = np.abs(refatoms.positions - atoms.positions).max()
    print(f'errs e={eerr} f={ferr} pos={perr}')
    assert eerr < 1e-11, eerr
    assert ferr < 1e-11, ferr
    assert perr < 1e-11, perr


def run_normal():
    atoms = getatoms()
    atoms.calc = EMT()
    with BFGS(atoms) as opt:
        opt.run()
    return atoms


def run_client(port, unixsocket):
    atoms = getatoms()
    atoms.calc = EMT()

    try:
        with open('client.log', 'w') as fd:
            client = SocketClient(log=fd, port=port,
                                  unixsocket=unixsocket,
                                  timeout=timeout)
            client.run(atoms, use_stress=False)
    except BrokenPipeError:
        # I think we can find a way to close sockets so as not to get an
        # error, but presently things are not like that.
        pass


def launch_client_thread(port, unixsocket):
    thread = threading.Thread(target=run_client, args=(port, unixsocket))
    thread.start()
    return thread


unix_only = pytest.mark.skipif(os.name != 'posix',
                               reason='requires unix platform')


@pytest.mark.optimize()
@pytest.mark.parametrize('sockettype', [
    'inet',
    pytest.param('unix', marks=unix_only),
])
def test_ipi_protocol(sockettype, testdir):
    try:
        run_server(sockettype=sockettype)
    except OSError as err:
        # The AppVeyor CI tests sometimes fail when we try to open sockets on
        # computers where this is forbidden.  For now we will simply skip
        # this test when that happens:
        if 'forbidden by its access permissions' in err.strerror:
            pytest.skip(err.strerror)
        else:
            raise



"""Test suit for the CP2K ASE calulator.

http://www.cp2k.org
Author: Ole Schuett <ole.schuett@mat.ethz.ch>
"""

import subprocess

import numpy as np
import pytest

from ase import io
from ase.build import molecule
from ase.calculators.calculator import compare_atoms
from ase.io.cp2k import iread_cp2k_dcd

inp = """\
&MOTION
  &PRINT
    &TRAJECTORY SILENT
      FORMAT DCD_ALIGNED_CELL
    &END TRAJECTORY
  &END PRINT
  &MD
    STEPS 5
  &END MD
&END MOTION
&GLOBAL
  RUN_TYPE MD
&END GLOBAL
"""


@pytest.fixture()
def cp2k_main(testconfig):
    try:
        return testconfig.parser['cp2k']['cp2k_main']
    except KeyError:
        pytest.skip("""\
Missing cp2k configuration.  Requires:

    [cp2k]
    cp2k_main = /path/to/cp2k
""")

# XXX multi-line skip messages are ugly since they mess up the
# pytest -r s listing.  Should we write a more terse message?
# Let's do it when calculator factory reporting is user friendly.


@pytest.mark.calculator_lite()
@pytest.mark.calculator('cp2k')
def test_dcd(factory, cp2k_main):
    calc = factory.calc(label='test_dcd', inp=inp)
    h2 = molecule('H2', calculator=calc)
    h2.center(vacuum=2.0)
    h2.set_pbc(True)
    energy = h2.get_potential_energy()
    assert energy is not None
    subprocess.check_call([cp2k_main, '-i', 'test_dcd.inp', '-o',
                           'test_dcd.out'])
    h2_end = io.read('test_dcd-pos-1.dcd')
    assert (h2_end.symbols == 'X').all()
    traj = io.read('test_dcd-pos-1.dcd', ref_atoms=h2,
                   index=slice(0, None), aligned=True)
    ioITraj = io.iread('test_dcd-pos-1.dcd', ref_atoms=h2,
                       index=slice(0, None), aligned=True)

    with open('test_dcd-pos-1.dcd', 'rb') as fd:
        itraj = iread_cp2k_dcd(fd, indices=slice(0, None),
                               ref_atoms=h2, aligned=True)
        for i, iMol in enumerate(itraj):
            ioIMol = next(ioITraj)
            assert compare_atoms(iMol, traj[i]) == []
            assert compare_atoms(iMol, ioIMol) == []
            assert iMol.get_pbc().all()

    traj = io.read('test_dcd-pos-1.dcd', ref_atoms=h2, index=slice(0, None))
    pbc = [mol.get_pbc() for mol in traj]
    assert not np.any(pbc)



import shutil
from math import sqrt
from os.path import getsize
from pathlib import Path

import numpy as np
import pytest

from ase.build import bulk
from ase.calculators.emt import EMT
from ase.filters import FrechetCellFilter
from ase.io.trajectory import Trajectory
from ase.optimize import BFGS, BFGSLineSearch, CellAwareBFGS


def params(opt):
    run_params = dict(fmax=0.005)
    opt_params = {}

    if opt is CellAwareBFGS:
        run_params.update(dict(smax=0.0005))

    if opt is BFGS:
        opt_params = {"append_trajectory": True}

    return run_params, opt_params


def opt_filter_atoms(opt, trajectory, restart, opt_params):
    atoms = bulk("Au")
    atoms *= 2
    atoms.rattle(stdev=0.005, seed=1)

    # if restarting, we take atoms from traj
    if Path(trajectory).is_file():
        with Trajectory(trajectory, 'r') as traj:
            atoms = traj[-1]

    atoms.calc = EMT()

    opt_relax = opt(
        FrechetCellFilter(atoms, exp_cell_factor=1.0),
        alpha=70,
        trajectory=trajectory,
        restart=restart,
        **opt_params,
    )

    return opt_relax


def fragile_optimizer(opt, trajectory, restart, run_kwargs, opt_params):
    break_now = 2

    with opt_filter_atoms(
        opt=opt, trajectory=trajectory, restart=restart, opt_params=opt_params
    ) as fragile_init:
        if isinstance(fragile_init, CellAwareBFGS):
            smax = run_kwargs.pop('smax')
            fragile_init.smax = smax
        for idx, _ in enumerate(fragile_init.irun(**run_kwargs)):
            if idx == break_now:
                break
        else:
            raise RuntimeError(
                'Fragile Optimizer did not break. Check if nsteps is to large.'
            )

    # pick up where we left off, assert we have written the files, and they
    # contain data. We check this here since these files are required in
    # order to properly restart.
    assert fragile_init.nsteps == break_now
    assert Path(restart).is_file() and Path(trajectory).is_file()
    assert all(size != 0 for size in [getsize(restart), getsize(trajectory)])

    if not opt_params.get('append_trajectory', False):
        shutil.copy(trajectory, trajectory + '.orig')

    with opt_filter_atoms(
        opt=opt, trajectory=trajectory, restart=restart, opt_params=opt_params
    ) as fragile_restart:
        if isinstance(fragile_restart, CellAwareBFGS):
            fragile_restart.smax = smax
        fragile_restart.run(**run_kwargs)

    return fragile_init, fragile_restart


@pytest.mark.parametrize(
    "opt", [BFGS, CellAwareBFGS, pytest.param(
        BFGSLineSearch, marks=pytest.mark.xfail(
            reason='Restart is absolutely broken and does not work. orig_cell '
                   'is not stored in output'))])
def test_optimizers_restart(testdir, opt):
    restart_filename = f"restart_{opt.__name__}.dat"
    trajectory_filename = f"{opt.__name__}.traj"
    run_kwargs, opt_params = params(opt)

    # single run
    with opt_filter_atoms(
        opt=opt,
        trajectory="single_" + trajectory_filename,
        restart="single_" + restart_filename,
        opt_params=opt_params,
    ) as single:
        single.run(**run_kwargs)

    # fragile restart
    fragile_init, fragile_restart = fragile_optimizer(
        opt=opt,
        trajectory="fragile_" + trajectory_filename,
        restart="fragile_" + restart_filename,
        run_kwargs=run_kwargs,
        opt_params=opt_params,
    )

    assert single.nsteps == fragile_init.nsteps + fragile_restart.nsteps

    single_traj = read_traj("single_" + trajectory_filename)
    fragile_traj = read_traj("fragile_" + trajectory_filename)

    if not opt_params.get('append_trajectory', False):
        fragile_traj_og = read_traj("fragile_" + trajectory_filename + '.orig')
        count = len(fragile_traj_og) - 1
        for traj_idx in fragile_traj:
            if traj_idx['step'] == 0:
                continue
            traj_idx['step'] += count
            fragile_traj_og.append(traj_idx)
        fragile_traj = fragile_traj_og

    # last step of init == first step of restart == single run break_now step
    # last step of restart == last step of single run
    for f_traj, s_traj in zip(fragile_traj, single_traj):
        for f, s in zip(f_traj, s_traj):
            assert np.allclose(f_traj[f], s_traj[s])


def read_traj(file: str):
    data = []

    with Trajectory(file, "r") as traj:
        for idx, atoms in enumerate(traj):

            pos = atoms.get_positions()
            forces = atoms.calc.results["forces"]
            stress = atoms.calc.results["stress"]
            energy = atoms.calc.results["energy"]

            fmax = sqrt((forces**2).sum(axis=1).max())
            smax = abs(stress).max()

            tmp = {
                "step": idx,
                "energy": energy,
                "position": pos,
                "forces": forces,
                "fmax": fmax,
                "smax": smax,
            }
            data.append(tmp)
        traj.close()

    return data



import pytest

import ase
from ase.build import bulk
from ase.calculators.lj import LennardJones
from ase.filters import Filter, FrechetCellFilter, UnitCellFilter
from ase.optimize.precon import Exp, PreconLBFGS


# @pytest.mark.skip('FAILS WITH PYAMG')
@pytest.mark.optimize()
@pytest.mark.slow()
# Ignore UserWarning by failure of Armijo line search in PreconLBFGS
@pytest.mark.filterwarnings('ignore::UserWarning')
@pytest.mark.parametrize("filter_cls, tol", [
    (UnitCellFilter, 1e-3),
    # FrechetCellFilter allows relaxing to lower tolerance
    (FrechetCellFilter, 1e-7),
])
def test_precon(filter_cls, tol: float):
    cu0: ase.Atoms = bulk("Cu") * (2, 2, 2)
    lj = LennardJones(sigma=cu0.get_distance(0, 1))

    ratio = 1.2
    cu = cu0.copy()
    cu.set_cell(ratio * cu.get_cell())
    cu.calc = lj
    filter: Filter = filter_cls(cu, constant_volume=True)
    opt = PreconLBFGS(filter, precon=Exp(mu=1.0, mu_c=1.0))
    opt.run(fmax=1e-3)
    assert abs(cu.get_volume() / cu0.get_volume() - ratio**3) < tol



# Stripped (minimal) version of nomad entry with 3 images.
# The images are actually identical for some reason, but we want to be sure
# that they are extracted correctly.
from ase.io import iread


def test_nomad(datadir):

    path = datadir / 'nomad-images.nomad-json'
    images = list(iread(path))
    assert len(images) == 3

    for atoms in images:
        assert all(atoms.pbc)
        assert (atoms.cell > 0).sum() == 3
        assert atoms.get_chemical_formula() == 'As24Sr32'


# Code for cleaning up nomad files so their size is reasonable for inclusion
# in test suite:
"""
ourkeys = {'section_run', 'section_system', 'name', 'atom_species',
           'atom_positions', 'flatData', 'uri', 'gIndex',
           'configuration_periodic_dimensions', 'lattice_vectors'}

includekeys = lambda k: k in ourkeys

fname = ...
with open(fname) as fd:
    d = read(fd, includekeys=includekeys)
print(json.dumps(d))
"""



import numpy as np
import pytest

from ase.build import molecule
from ase.calculators.siesta.siesta_lrtddft import SiestaLRTDDFT


def test_siesta_lrtddft(siesta_factory):

    pynao = pytest.importorskip('pynao')
    print("pynao version: ", pynao.__version__)

    # Define the systems
    ch4 = molecule('CH4')

    lrtddft = SiestaLRTDDFT(label="siesta", xc_code='LDA,PZ')

    # run siesta
    lrtddft.get_ground_state(ch4)

    freq = np.arange(0.0, 25.0, 0.5)
    pmat = lrtddft.get_polarizability(freq)
    assert pmat.size == 3 * 3 * freq.size



"""ASE Calculator for the ground state exciting DFT code.

Exciting calculator class in this file allow for writing exciting input
files using ASE Atoms object that allow for the compiled exciting binary
to run DFT on the geometry/material defined in the Atoms object. Also gives
access to developer to a lightweight parser (lighter weight than NOMAD or
the exciting parser in the exciting repository) to capture ground state
properties.

Note: excitingtools must be installed using `pip install excitingtools` to
use this calculator.
"""

from os import PathLike
from pathlib import Path
from typing import Any, Mapping

import ase.io.exciting
from ase.calculators.calculator import PropertyNotImplementedError
from ase.calculators.exciting.runner import (
    SimpleBinaryRunner,
    SubprocessRunResults,
)
from ase.calculators.genericfileio import (
    BaseProfile,
    CalculatorTemplate,
    GenericFileIOCalculator,
)


class ExcitingProfile(BaseProfile):
    """Defines all quantities that are configurable for a given machine.

    Follows the generic pattern BUT currently not used by our calculator as:
       * species_path is part of the input file in exciting.
       * OnlyTypo fix part of the profile used in the base class is the run
         method, which is part of the BinaryRunner class.
    """
    configvars = {'species_path'}

    def __init__(self, command, species_path=None, **kwargs):
        super().__init__(command, **kwargs)

        self.species_path = species_path

    def version(self):
        """Return exciting version."""
        # TARP No way to get the version for the binary in use
        return

    # Machine specific config files in the config
    # species_file goes in the config
    # binary file in the config.
    # options for that, parallel info dictionary.
    # Number of threads and stuff like that.

    def get_calculator_command(self, input_file):
        """Returns command to run binary as a list of strings."""
        # input_file unused for exciting, it looks for input.xml in run
        # directory.
        if input_file is None:
            return []
        else:
            return [str(input_file)]


class ExcitingGroundStateTemplate(CalculatorTemplate):
    """Template for Ground State Exciting Calculator

    Abstract methods inherited from the base class:
        * write_input
        * execute
        * read_results
    """

    parser = {'info.xml': ase.io.exciting.parse_output}
    output_names = list(parser)
    # Use frozenset since the CalculatorTemplate enforces it.
    implemented_properties = frozenset(['energy', 'forces'])
    _label = 'exciting'

    def __init__(self):
        """Initialise with constant class attributes.

        :param program_name: The DFT program, should always be exciting.
        :param implemented_properties: What properties should exciting
            calculate/read from output.
        """
        super().__init__('exciting', self.implemented_properties)
        self.errorname = f'{self._label}.err'

    @staticmethod
    def _require_forces(input_parameters):
        """Expect ASE always wants forces, enforce setting in input_parameters.

        :param input_parameters: exciting ground state input parameters, either
            as a dictionary or ExcitingGroundStateInput.
        :return: Ground state input parameters, with "compute
                forces" set to true.
        """
        from excitingtools import ExcitingGroundStateInput

        input_parameters = ExcitingGroundStateInput(input_parameters)
        input_parameters.tforce = True
        return input_parameters

    def write_input(
        self,
        profile: ExcitingProfile,  # ase test linter enforces method signatures
        # be consistent with the
        # abstract method that it implements
        directory: PathLike,
        atoms: ase.Atoms,
        parameters: dict,
        properties=None,
    ):
        """Write an exciting input.xml file based on the input args.

        :param profile: an Exciting code profile
        :param directory: Directory in which to run calculator.
        :param atoms: ASE atoms object.
        :param parameters: exciting ground state input parameters, in a
            dictionary. Expect species_path, title and ground_state data,
            either in an object or as dict.
        :param properties: Base method's API expects the physical properties
            expected from a ground state calculation, for example energies
            and forces. For us this is not used.
        """
        # Create a copy of the parameters dictionary so we don't
        # modify the callers dictionary.
        parameters_dict = parameters
        assert set(parameters_dict.keys()) == {
            'title', 'species_path', 'ground_state_input',
            'properties_input'}, \
            'Keys should be defined by ExcitingGroundState calculator'
        file_name = Path(directory) / 'input.xml'
        species_path = parameters_dict.pop('species_path')
        title = parameters_dict.pop('title')
        # We can also pass additional parameters which are actually called
        # properties in the exciting input xml. We don't use this term
        # since ASE use properties to refer to results of a calculation
        # (e.g. force, energy).
        if 'properties_input' not in parameters_dict:
            parameters_dict['properties_input'] = None

        ase.io.exciting.write_input_xml_file(
            file_name=file_name, atoms=atoms,
            ground_state_input=parameters_dict['ground_state_input'],
            species_path=species_path, title=title,
            properties_input=parameters_dict['properties_input'])

    def execute(
            self, directory: PathLike,
            profile) -> SubprocessRunResults:
        """Given an exciting calculation profile, execute the calculation.

        :param directory: Directory in which to execute the calculator
            exciting_calculation: Base method `execute` expects a profile,
            however it is simply used to execute the program, therefore we
            just pass a SimpleBinaryRunner.
        :param profile: This name comes from the superclass CalculatorTemplate.
                It contains machine specific information to run the
                calculation.

        :return: Results of the subprocess.run command.
        """
        return profile.run(directory, f"{directory}/input.xml", None,
                           erorrfile=self.errorname)

    def read_results(self, directory: PathLike) -> Mapping[str, Any]:
        """Parse results from each ground state output file.

        Note we allow for the ability for there to be multiple output files.

        :param directory: Directory path to output file from exciting
            simulation.
        :return: Dictionary containing important output properties.
        """
        results = {}
        for file_name in self.output_names:
            full_file_path = Path(directory) / file_name
            result: dict = self.parser[file_name](full_file_path)
            results.update(result)
        return results

    def load_profile(self, cfg, **kwargs):
        """ExcitingProfile can be created via a config file.

        Alternative to this method the profile can be created with it's
        init method. This method allows for more settings to be passed.
        """
        return ExcitingProfile.from_config(cfg, self.name, **kwargs)


class ExcitingGroundStateResults:
    """Exciting Ground State Results."""

    def __init__(self, results: dict) -> None:
        self.results = results
        self.final_scl_iteration = list(results['scl'].keys())[-1]

    def total_energy(self) -> float:
        """Return total energy of system."""
        # TODO(Alex) We should a common list of keys somewhere
        # such that parser -> results -> getters are consistent
        return float(
            self.results['scl'][self.final_scl_iteration]['Total energy']
        )

    def band_gap(self) -> float:
        """Return the estimated fundamental gap from the exciting sim."""
        return float(
            self.results['scl'][self.final_scl_iteration][
                'Estimated fundamental gap'
            ]
        )

    def forces(self):
        """Return forces present on the system.

        Currently, not all exciting simulations return forces. We leave this
        definition for future revisions.
        """
        raise PropertyNotImplementedError

    def stress(self):
        """Get the stress on the system.

        Right now exciting does not yet calculate the stress on the system so
        this won't work for the time being.
        """
        raise PropertyNotImplementedError


class ExcitingGroundStateCalculator(GenericFileIOCalculator):
    """Class for the ground state calculation.

    :param runner: Binary runner that will execute an exciting calculation and
        return a result.
    :param ground_state_input: dictionary of ground state settings for example
        {'rgkmax': 8.0, 'autormt': True} or an object of type
        ExcitingGroundStateInput.
    :param directory: Directory in which to run the job.
    :param species_path: Path to the location of exciting's species files.
    :param title: job name written to input.xml

    :return: Results returned from running the calculate method.


    Typical usage:

    gs_calculator = ExcitingGroundState(runner, ground_state_input)

    results: ExcitingGroundStateResults = gs_calculator.calculate(
            atoms: Atoms)
    """

    def __init__(
        self,
        *,
        runner: SimpleBinaryRunner,
        ground_state_input,
        directory='./',
        species_path='./',
        title='ASE-generated input',
    ):
        self.runner = runner
        # Package data to be passed to
        # ExcitingGroundStateTemplate.write_input(..., input_parameters, ...)
        # Structure not included, as it's passed when one calls .calculate
        # method directly
        self.exciting_inputs = {
            'title': title,
            'species_path': species_path,
            'ground_state_input': ground_state_input,
        }
        self.directory = Path(directory)

        # GenericFileIOCalculator expects a `profile`
        # containing machine-specific settings, however, in exciting's case,
        # the species file are defined in the input XML (hence passed in the
        # parameters argument) and the only other machine-specific setting is
        # the BinaryRunner. Furthermore, in GenericFileIOCalculator.calculate,
        # profile is only used to provide a run method. We therefore pass the
        # BinaryRunner in the place of a profile.
        super().__init__(
            profile=runner,
            template=ExcitingGroundStateTemplate(),
            directory=directory,
            parameters=self.exciting_inputs,
        )



# type: ignore
import numpy as np

from ase.build import molecule
from ase.calculators.turbomole import Turbomole


def test_turbomole_h2o(turbomole_factory):
    mol = molecule('H2O')

    params = {
        'title': 'water',
        'task': 'geometry optimization',
        'use redundant internals': True,
        'basis set name': 'def2-SV(P)',
        'total charge': 0,
        'multiplicity': 1,
        'use dft': True,
        'density functional': 'b3-lyp',
        'use resolution of identity': True,
        'ri memory': 1000,
        'force convergence': 0.001,
        'geometry optimization iterations': 50,
        'scf iterations': 100
    }

    calc = Turbomole(**params)
    mol.calc = calc
    calc.calculate(mol)
    assert calc.converged

    # use the get_property() method
    energy = calc.get_property('energy', mol, False)
    assert energy is not False
    assert np.isclose(energy, -2076.286138, atol=1e-5)
    forces = calc.get_property('forces', mol, False)
    assert forces is not False
    assert np.linalg.norm(forces) < 0.01
    dipole = np.linalg.norm(calc.get_property('dipole', mol, False))
    assert np.isclose(dipole, 0.448, rtol=0.01)

    # use the get_results() method
    results = calc.get_results()
    print(results['molecular orbitals'])

    # use the __getitem__() method
    print(calc['results']['molecular orbitals'])
    print(calc['results']['geometry optimization history'])

    # perform a normal mode calculation with the optimized structure

    params.update({
        'task': 'normal mode analysis',
        'density convergence': 1.0e-7
    })

    calc = Turbomole(**params)
    mol.calc = calc
    calc.calculate(mol)

    spectrum = calc['results']['vibrational spectrum']
    freq_ref = [0, 0, 0, 0, 0, 0, 1633, 3637, 3745]
    for s in spectrum:
        for freq, mode in zip(freq_ref, range(1, 7)):
            if s['mode number'] == mode:
                assert np.isclose(s['frequency']['value'], freq, rtol=0.05)

    print(calc.todict(skip_default=False))


def test_turbomole_h2o_parsing_issue(turbomole_factory):
    """setting the damping in define causes a parsing issue during restart"""
    mol = molecule('H2O')
    params = {
        'multiplicity': 1,
        'initial damping': 2.0,
        'damping adjustment step': 0.5,
        'minimal damping': 1.0
    }
    mol.calc = Turbomole(**params)
    mol.get_potential_energy(mol)
    Turbomole(restart=True)



from ase.build import bulk
from ase.calculators.emt import EMT
from ase.calculators.loggingcalc import LoggingCalculator


def test_loggingcalc(tmp_path, figure):
    atoms = bulk('Au', orthorhombic=True)
    calc = LoggingCalculator(EMT())
    atoms.calc = calc

    for _ in range(4):
        atoms.get_forces()
        atoms.get_potential_energy()
        atoms.positions[0, 0] += 0.1

    # This is kind of improper testing.
    # Should split plotting away from processing so we can
    # test the processing.
    calc.tabulate()
    calc.plot()



"""Module for calculating phonons of periodic systems."""

import warnings
from math import pi, sqrt
from pathlib import Path

import numpy as np
import numpy.fft as fft
import numpy.linalg as la

import ase
import ase.units as units
from ase.dft import monkhorst_pack
from ase.io.trajectory import Trajectory
from ase.parallel import world
from ase.utils import deprecated
from ase.utils.filecache import MultiFileJSONCache


class Displacement:
    """Abstract base class for phonon and el-ph supercell calculations.

    Both phonons and the electron-phonon interaction in periodic systems can be
    calculated with the so-called finite-displacement method where the
    derivatives of the total energy and effective potential are obtained from
    finite-difference approximations, i.e. by displacing the atoms. This class
    provides the required functionality for carrying out the calculations for
    the different displacements in its ``run`` member function.

    Derived classes must overwrite the ``__call__`` member function which is
    called for each atomic displacement.

    """

    def __init__(self, atoms, calc=None, supercell=(1, 1, 1), name=None,
                 delta=0.01, center_refcell=False, comm=None):
        """Init with an instance of class ``Atoms`` and a calculator.

        Parameters:

        atoms: Atoms object
            The atoms to work on.
        calc: Calculator
            Calculator for the supercell calculation.
        supercell: tuple
            Size of supercell given by the number of repetitions (l, m, n) of
            the small unit cell in each direction.
        name: str
            Base name to use for files.
        delta: float
            Magnitude of displacement in Ang.
        center_refcell: bool
            Reference cell in which the atoms will be displaced. If False, then
            corner cell in supercell is used. If True, then cell in the center
            of the supercell is used.
        comm: communicator
            MPI communicator for the phonon calculation.
            Default is to use world.
        """

        # Store atoms and calculator
        self.atoms = atoms
        self.calc = calc

        # Displace all atoms in the unit cell by default
        self.indices = np.arange(len(atoms))
        self.name = name
        self.delta = delta
        self.center_refcell = center_refcell
        self.supercell = supercell

        if comm is None:
            comm = world
        self.comm = comm

        self.cache = MultiFileJSONCache(self.name)

    def define_offset(self):        # Reference cell offset

        if not self.center_refcell:
            # Corner cell
            self.offset = 0
        else:
            # Center cell
            N_c = self.supercell
            self.offset = (N_c[0] // 2 * (N_c[1] * N_c[2]) +
                           N_c[1] // 2 * N_c[2] +
                           N_c[2] // 2)
        return self.offset

    @property
    @ase.utils.deprecated('Please use phonons.supercell instead of .N_c')
    def N_c(self):
        return self._supercell

    @property
    def supercell(self):
        return self._supercell

    @supercell.setter
    def supercell(self, supercell):
        assert len(supercell) == 3
        self._supercell = tuple(supercell)
        self.define_offset()
        self._lattice_vectors_array = self.compute_lattice_vectors()

    @ase.utils.deprecated('Please use phonons.compute_lattice_vectors()'
                          ' instead of .lattice_vectors()')
    def lattice_vectors(self):
        return self.compute_lattice_vectors()

    def compute_lattice_vectors(self):
        """Return lattice vectors for cells in the supercell."""
        # Lattice vectors -- ordered as illustrated in class docstring

        # Lattice vectors relevative to the reference cell
        R_cN = np.indices(self.supercell).reshape(3, -1)
        N_c = np.array(self.supercell)[:, np.newaxis]
        if self.offset == 0:
            R_cN += N_c // 2
            R_cN %= N_c
        R_cN -= N_c // 2
        return R_cN

    def __call__(self, *args, **kwargs):
        """Member function called in the ``run`` function."""

        raise NotImplementedError("Implement in derived classes!.")

    def set_atoms(self, atoms):
        """Set the atoms to vibrate.

        Parameters:

        atoms: list
            Can be either a list of strings, ints or ...

        """

        assert isinstance(atoms, list)
        assert len(atoms) <= len(self.atoms)

        if isinstance(atoms[0], str):
            assert np.all([isinstance(atom, str) for atom in atoms])
            sym_a = self.atoms.get_chemical_symbols()
            # List for atomic indices
            indices = []
            for type in atoms:
                indices.extend([a for a, atom in enumerate(sym_a)
                                if atom == type])
        else:
            assert np.all([isinstance(atom, int) for atom in atoms])
            indices = atoms

        self.indices = indices

    def _eq_disp(self):
        return self._disp(0, 0, 0)

    def _disp(self, a, i, step):
        from ase.vibrations.vibrations import Displacement as VDisplacement
        return VDisplacement(a, i, np.sign(step), abs(step), self)

    def run(self):
        """Run the calculations for the required displacements.

        This will do a calculation for 6 displacements per atom, +-x, +-y, and
        +-z. Only those calculations that are not already done will be
        started. Be aware that an interrupted calculation may produce an empty
        file (ending with .json), which must be deleted before restarting the
        job. Otherwise the calculation for that displacement will not be done.

        """

        # Atoms in the supercell -- repeated in the lattice vector directions
        # beginning with the last
        atoms_N = self.atoms * self.supercell

        # Set calculator if provided
        assert self.calc is not None, "Provide calculator in __init__ method"
        atoms_N.calc = self.calc

        # Do calculation on equilibrium structure
        eq_disp = self._eq_disp()
        with self.cache.lock(eq_disp.name) as handle:
            if handle is not None:
                output = self.calculate(atoms_N, eq_disp)
                handle.save(output)

        # Positions of atoms to be displaced in the reference cell
        natoms = len(self.atoms)
        offset = natoms * self.offset
        pos = atoms_N.positions[offset: offset + natoms].copy()

        # Loop over all displacements
        for a in self.indices:
            for i in range(3):
                for sign in [-1, 1]:
                    disp = self._disp(a, i, sign)
                    with self.cache.lock(disp.name) as handle:
                        if handle is None:
                            continue
                        try:
                            atoms_N.positions[offset + a, i] = \
                                pos[a, i] + sign * self.delta

                            result = self.calculate(atoms_N, disp)
                            handle.save(result)
                        finally:
                            # Return to initial positions
                            atoms_N.positions[offset + a, i] = pos[a, i]

        self.comm.barrier()

    def clean(self):
        """Delete generated files."""
        if self.comm.rank == 0:
            nfiles = self._clean()
        else:
            nfiles = 0
        self.comm.barrier()
        return nfiles

    def _clean(self):
        name = Path(self.name)

        nfiles = 0
        if name.is_dir():
            for fname in name.iterdir():
                fname.unlink()
                nfiles += 1
            name.rmdir()
        return nfiles


class Phonons(Displacement):
    r"""Class for calculating phonon modes using the finite displacement method.

    The matrix of force constants is calculated from the finite difference
    approximation to the first-order derivative of the atomic forces as::

                            2             nbj   nbj
                nbj        d E           F-  - F+
               C     = ------------ ~  -------------  ,
                mai     dR   dR          2 * delta
                          mai  nbj

    where F+/F- denotes the force in direction j on atom nb when atom ma is
    displaced in direction +i/-i. The force constants are related by various
    symmetry relations. From the definition of the force constants it must
    be symmetric in the three indices mai::

                nbj    mai         bj        ai
               C    = C      ->   C  (R ) = C  (-R )  .
                mai    nbj         ai  n     bj   n

    As the force constants can only depend on the difference between the m and
    n indices, this symmetry is more conveniently expressed as shown on the
    right hand-side.

    The acoustic sum-rule::

                           _ _
                aj         \    bj
               C  (R ) = -  )  C  (R )
                ai  0      /__  ai  m
                          (m, b)
                            !=
                          (0, a)

    Ordering of the unit cells illustrated here for a 1-dimensional system (in
    case ``refcell=None`` in constructor!):

    ::

               m = 0        m = 1        m = -2        m = -1
           -----------------------------------------------------
           |            |            |            |            |
           |        * b |        *   |        *   |        *   |
           |            |            |            |            |
           |   * a      |   *        |   *        |   *        |
           |            |            |            |            |
           -----------------------------------------------------

    Example:

    >>> from ase.build import bulk
    >>> from ase.phonons import Phonons
    >>> from gpaw import GPAW, FermiDirac

    >>> atoms = bulk('Si', 'diamond', a=5.4)
    >>> calc = GPAW(mode='fd',
    ...             kpts=(5, 5, 5),
    ...             h=0.2,
    ...             occupations=FermiDirac(0.))
    >>> ph = Phonons(atoms, calc, supercell=(5, 5, 5))
    >>> ph.run()
    >>> ph.read(method='frederiksen', acoustic=True)

    """

    def __init__(self, *args, **kwargs):
        """Initialize with base class args and kwargs."""

        if 'name' not in kwargs:
            kwargs['name'] = "phonon"

        self.deprecate_refcell(kwargs)

        Displacement.__init__(self, *args, **kwargs)

        # Attributes for force constants and dynamical matrix in real space
        self.C_N = None  # in units of eV / Ang**2
        self.D_N = None  # in units of eV / Ang**2 / amu

        # Attributes for born charges and static dielectric tensor
        self.Z_avv = None
        self.eps_vv = None

    @staticmethod
    def deprecate_refcell(kwargs: dict):
        if 'refcell' in kwargs:
            warnings.warn('Keyword refcell of Phonons is deprecated.'
                          'Please use center_refcell (bool)', FutureWarning)
            kwargs['center_refcell'] = bool(kwargs['refcell'])
            kwargs.pop('refcell')

        return kwargs

    def __call__(self, atoms_N):
        """Calculate forces on atoms in supercell."""
        return atoms_N.get_forces()

    def calculate(self, atoms_N, disp):
        forces = self(atoms_N)
        return {'forces': forces}

    def check_eq_forces(self):
        """Check maximum size of forces in the equilibrium structure."""

        eq_disp = self._eq_disp()
        feq_av = self.cache[eq_disp.name]['forces']

        fmin = feq_av.min()
        fmax = feq_av.max()
        i_min = np.where(feq_av == fmin)
        i_max = np.where(feq_av == fmax)

        return fmin, fmax, i_min, i_max

    @deprecated('Current implementation of non-analytical correction is '
                'likely incorrect, see '
                'https://gitlab.com/ase/ase/-/issues/941')
    def read_born_charges(self, name='born', neutrality=True):
        r"""Read Born charges and dieletric tensor from JSON file.

        The charge neutrality sum-rule::

                   _ _
                   \    a
                    )  Z   = 0
                   /__  ij
                    a

        Parameters:

        neutrality: bool
            Restore charge neutrality condition on calculated Born effective
            charges.
        name: str
            Key used to identify the file with Born charges for the unit cell
            in the JSON cache.

        .. deprecated:: 3.22.1
            Current implementation of non-analytical correction is likely
            incorrect, see :issue:`941`
        """

        # Load file with Born charges and dielectric tensor for atoms in the
        # unit cell
        Z_avv, eps_vv = self.cache[name]

        # Neutrality sum-rule
        if neutrality:
            Z_mean = Z_avv.sum(0) / len(Z_avv)
            Z_avv -= Z_mean

        self.Z_avv = Z_avv[self.indices]
        self.eps_vv = eps_vv

    def read(self, method='Frederiksen', symmetrize=3, acoustic=True,
             cutoff=None, born=False, **kwargs):
        """Read forces from json files and calculate force constants.

        Extra keyword arguments will be passed to ``read_born_charges``.

        Parameters:

        method: str
            Specify method for evaluating the atomic forces.
        symmetrize: int
            Symmetrize force constants (see doc string at top) when
            ``symmetrize != 0`` (default: 3). Since restoring the acoustic sum
            rule breaks the symmetry, the symmetrization must be repeated a few
            times until the changes a insignificant. The integer gives the
            number of iterations that will be carried out.
        acoustic: bool
            Restore the acoustic sum rule on the force constants.
        cutoff: None or float
            Zero elements in the dynamical matrix between atoms with an
            interatomic distance larger than the cutoff.
        born: bool
            Read in Born effective charge tensor and high-frequency static
            dielelctric tensor from file.

        """

        method = method.lower()
        assert method in ['standard', 'frederiksen']
        if cutoff is not None:
            cutoff = float(cutoff)

        # Read Born effective charges and optical dielectric tensor
        if born:
            self.read_born_charges(**kwargs)

        # Number of atoms
        natoms = len(self.indices)
        # Number of unit cells
        N = np.prod(self.supercell)
        # Matrix of force constants as a function of unit cell index in units
        # of eV / Ang**2
        C_xNav = np.empty((natoms * 3, N, natoms, 3), dtype=float)

        # Loop over all atomic displacements and calculate force constants
        for i, a in enumerate(self.indices):
            for j, v in enumerate('xyz'):
                # Atomic forces for a displacement of atom a in direction v
                # basename = '%s.%d%s' % (self.name, a, v)
                basename = '%d%s' % (a, v)
                fminus_av = self.cache[basename + '-']['forces']
                fplus_av = self.cache[basename + '+']['forces']

                if method == 'frederiksen':
                    fminus_av[a] -= fminus_av.sum(0)
                    fplus_av[a] -= fplus_av.sum(0)

                # Finite difference derivative
                C_av = fminus_av - fplus_av
                C_av /= 2 * self.delta

                # Slice out included atoms
                C_Nav = C_av.reshape((N, len(self.atoms), 3))[:, self.indices]
                index = 3 * i + j
                C_xNav[index] = C_Nav

        # Make unitcell index the first and reshape
        C_N = C_xNav.swapaxes(0, 1).reshape((N,) + (3 * natoms, 3 * natoms))

        # Cut off before symmetry and acoustic sum rule are imposed
        if cutoff is not None:
            self.apply_cutoff(C_N, cutoff)

        # Symmetrize force constants
        if symmetrize:
            for _ in range(symmetrize):
                # Symmetrize
                C_N = self.symmetrize(C_N)
                # Restore acoustic sum-rule
                if acoustic:
                    self.acoustic(C_N)
                else:
                    break

        # Store force constants and dynamical matrix
        self.C_N = C_N
        self.D_N = C_N.copy()

        # Add mass prefactor
        m_a = self.atoms.get_masses()
        self.m_inv_x = np.repeat(m_a[self.indices]**-0.5, 3)
        M_inv = np.outer(self.m_inv_x, self.m_inv_x)
        for D in self.D_N:
            D *= M_inv

    def symmetrize(self, C_N):
        """Symmetrize force constant matrix."""

        # Number of atoms
        natoms = len(self.indices)
        # Number of unit cells
        N = np.prod(self.supercell)

        # Reshape force constants to (l, m, n) cell indices
        C_lmn = C_N.reshape(self.supercell + (3 * natoms, 3 * natoms))

        # Shift reference cell to center index
        if self.offset == 0:
            C_lmn = fft.fftshift(C_lmn, axes=(0, 1, 2)).copy()
        # Make force constants symmetric in indices -- in case of an even
        # number of unit cells don't include the first cell
        i, j, k = 1 - np.asarray(self.supercell) % 2
        C_lmn[i:, j:, k:] *= 0.5
        C_lmn[i:, j:, k:] += \
            C_lmn[i:, j:, k:][::-1, ::-1, ::-1].transpose(0, 1, 2, 4, 3).copy()
        if self.offset == 0:
            C_lmn = fft.ifftshift(C_lmn, axes=(0, 1, 2)).copy()

        # Change to single unit cell index shape
        C_N = C_lmn.reshape((N, 3 * natoms, 3 * natoms))

        return C_N

    def acoustic(self, C_N):
        """Restore acoustic sumrule on force constants."""

        # Number of atoms
        natoms = len(self.indices)
        # Copy force constants
        C_N_temp = C_N.copy()

        # Correct atomic diagonals of R_m = (0, 0, 0) matrix
        for C in C_N_temp:
            for a in range(natoms):
                for a_ in range(natoms):
                    C_N[self.offset,
                        3 * a: 3 * a + 3,
                        3 * a: 3 * a + 3] -= C[3 * a: 3 * a + 3,
                                               3 * a_: 3 * a_ + 3]

    def apply_cutoff(self, D_N, r_c):
        """Zero elements for interatomic distances larger than the cutoff.

        Parameters:

        D_N: ndarray
            Dynamical/force constant matrix.
        r_c: float
            Cutoff in Angstrom.

        """

        # Number of atoms and primitive cells
        natoms = len(self.indices)
        N = np.prod(self.supercell)
        # Lattice vectors
        R_cN = self._lattice_vectors_array
        # Reshape matrix to individual atomic and cartesian dimensions
        D_Navav = D_N.reshape((N, natoms, 3, natoms, 3))

        # Cell vectors
        cell_vc = self.atoms.cell.transpose()
        # Atomic positions in reference cell
        pos_av = self.atoms.get_positions()

        # Zero elements with a distance to atoms in the reference cell
        # larger than the cutoff
        for n in range(N):
            # Lattice vector to cell
            R_v = np.dot(cell_vc, R_cN[:, n])
            # Atomic positions in cell
            posn_av = pos_av + R_v
            # Loop over atoms and zero elements
            for i, a in enumerate(self.indices):
                dist_a = np.sqrt(np.sum((pos_av[a] - posn_av)**2, axis=-1))
                # Atoms where the distance is larger than the cufoff
                i_a = dist_a > r_c  # np.where(dist_a > r_c)
                # Zero elements
                D_Navav[n, i, :, i_a, :] = 0.0

    def get_force_constant(self):
        """Return matrix of force constants."""

        assert self.C_N is not None
        return self.C_N

    def get_band_structure(self, path, modes=False, born=False, verbose=True):
        """Calculate and return the phonon band structure.

        This method computes the phonon band structure for a given path
        in reciprocal space. It is a wrapper around the internal
        `band_structure` method of the `Phonons` class. The method can
        optionally calculate and return phonon modes.

        Frequencies and modes are in units of eV and 1/sqrt(amu),
        respectively.

        Parameters:

        path : BandPath object
            The BandPath object defining the path in the reciprocal
            space over which the phonon band structure is calculated.
        modes : bool, optional
            If True, phonon modes will also be calculated and returned.
            Defaults to False.
        born : bool, optional
            If True, includes the effect of Born effective charges in
            the phonon calculations.
            Defaults to False.
        verbose : bool, optional
            If True, enables verbose output during the calculation.
            Defaults to True.

        Returns:

        BandStructure or tuple of (BandStructure, ndarray)
            If ``modes`` is False, returns a ``BandStructure`` object
            containing the phonon band structure. If ``modes`` is True,
            returns a tuple, where the first element is the
            ``BandStructure`` object and the second element is an ndarray
            of phonon modes.

            If modes are returned, the array is of shape
            (k-point, bands, atoms, 3) and the norm-squared of the mode
            is `1 / m_{eff}`, where `m_{eff}` is the effective mass of the
            mode.

        Example:

        >>> from ase.dft.kpoints import BandPath
        >>> path = BandPath(...)  # Define the band path
        >>> phonons = Phonons(...)
        >>> bs, modes = phonons.get_band_structure(path, modes=True)
        """
        result = self.band_structure(path.kpts,
                                     modes=modes,
                                     born=born,
                                     verbose=verbose)
        if modes:
            omega_kl, omega_modes = result
        else:
            omega_kl = result

        from ase.spectrum.band_structure import BandStructure
        bs = BandStructure(path, energies=omega_kl[None])

        # Return based on the modes flag
        return (bs, omega_modes) if modes else bs

    def compute_dynamical_matrix(self, q_scaled: np.ndarray, D_N: np.ndarray):
        """ Computation of the dynamical matrix in momentum space D_ab(q).
            This is a Fourier transform from real-space dynamical matrix D_N
            for a given momentum vector q.

        q_scaled: q vector in scaled coordinates.

        D_N: the dynamical matrix in real-space. It is necessary, at least
             currently, to provide this matrix explicitly (rather than use
             self.D_N) because this matrix is modified by the Born charges
             contributions and these modifications are momentum (q) dependent.

        Result:
            D(q): two-dimensional, complex-valued array of
                  shape=(3 * natoms, 3 * natoms).
        """
        # Evaluate fourier sum
        R_cN = self._lattice_vectors_array
        phase_N = np.exp(-2.j * pi * np.dot(q_scaled, R_cN))
        D_q = np.sum(phase_N[:, np.newaxis, np.newaxis] * D_N, axis=0)
        return D_q

    def band_structure(self, path_kc, modes=False, born=False, verbose=True):
        """Calculate phonon dispersion along a path in the Brillouin zone.

        The dynamical matrix at arbitrary q-vectors is obtained by Fourier
        transforming the real-space force constants. In case of negative
        eigenvalues (squared frequency), the corresponding negative frequency
        is returned.

        Frequencies and modes are in units of eV and 1/sqrt(amu),
        respectively.

        Parameters:

        path_kc: ndarray
            List of k-point coordinates (in units of the reciprocal lattice
            vectors) specifying the path in the Brillouin zone for which the
            dynamical matrix will be calculated.
        modes: bool
            Returns both frequencies and modes when True.
        born: bool
            Include non-analytic part given by the Born effective charges and
            the static part of the high-frequency dielectric tensor. This
            contribution to the force constant accounts for the splitting
            between the LO and TO branches for q -> 0.
        verbose: bool
            Print warnings when imaginary frequncies are detected.

        Returns:

        If modes=False: Array of energies

        If modes=True: Tuple of two arrays with energies and modes.
        """

        assert self.D_N is not None
        if born:
            assert self.Z_avv is not None
            assert self.eps_vv is not None

        # Dynamical matrix in real-space
        D_N = self.D_N

        # Lists for frequencies and modes along path
        omega_kl = []
        u_kl = []

        # Reciprocal basis vectors for use in non-analytic contribution
        reci_vc = 2 * pi * la.inv(self.atoms.cell)
        # Unit cell volume in Bohr^3
        vol = abs(la.det(self.atoms.cell)) / units.Bohr**3

        for q_c in path_kc:
            # Add non-analytic part
            if born:
                # q-vector in cartesian coordinates
                q_v = np.dot(reci_vc, q_c)
                # Non-analytic contribution to force constants in atomic units
                qdotZ_av = np.dot(q_v, self.Z_avv).ravel()
                C_na = (4 * pi * np.outer(qdotZ_av, qdotZ_av) /
                        np.dot(q_v, np.dot(self.eps_vv, q_v)) / vol)
                self.C_na = C_na / units.Bohr**2 * units.Hartree
                # Add mass prefactor and convert to eV / (Ang^2 * amu)
                M_inv = np.outer(self.m_inv_x, self.m_inv_x)
                D_na = C_na * M_inv / units.Bohr**2 * units.Hartree
                self.D_na = D_na
                D_N = self.D_N + D_na / np.prod(self.supercell)

            # if np.prod(self.N_c) == 1:
            #
            #     q_av = np.tile(q_v, len(self.indices))
            #     q_xx = np.vstack([q_av]*len(self.indices)*3)
            #     D_m += q_xx

            # Evaluate fourier sum
            D_q = self.compute_dynamical_matrix(q_c, D_N)

            if modes:
                omega2_l, u_xl = la.eigh(D_q, UPLO='U')
                # Sort eigenmodes according to eigenvalues (see below) and
                # multiply with mass prefactor.  This gives the eigenmode
                # (which is now not normalized!) with units 1/sqrt(amu).
                u_lx = (self.m_inv_x[:, np.newaxis] *
                        u_xl[:, omega2_l.argsort()]).T.copy()
                u_kl.append(u_lx.reshape((-1, len(self.indices), 3)))
            else:
                omega2_l = la.eigvalsh(D_q, UPLO='U')

            # Sort eigenvalues in increasing order
            omega2_l.sort()
            # Use dtype=complex to handle negative eigenvalues
            omega_l = np.sqrt(omega2_l.astype(complex))

            # Take care of imaginary frequencies
            if not np.all(omega2_l >= 0.):
                indices = np.where(omega2_l < 0)[0]

                if verbose:
                    print('WARNING, %i imaginary frequencies at '
                          'q = (% 5.2f, % 5.2f, % 5.2f) ; (omega_q =% 5.3e*i)'
                          % (len(indices), q_c[0], q_c[1], q_c[2],
                             omega_l[indices][0].imag))

                omega_l[indices] = -1 * np.sqrt(np.abs(omega2_l[indices].real))

            omega_kl.append(omega_l.real)

        # Conversion factor: sqrt(eV / Ang^2 / amu) -> eV
        s = units._hbar * 1e10 / sqrt(units._e * units._amu)
        omega_kl = s * np.asarray(omega_kl)

        if modes:
            return omega_kl, np.asarray(u_kl)

        return omega_kl

    def get_dos(self, kpts=(10, 10, 10), indices=None, verbose=True):
        """Return a phonon density of states.

        Parameters:

        kpts: tuple
            Shape of Monkhorst-Pack grid for sampling the Brillouin zone.
        indices: list
            If indices is not None, the amplitude-weighted atomic-partial
            DOS for the specified atoms will be calculated.
        verbose: bool
            Print warnings when imaginary frequncies are detected.

        Returns:
            A RawDOSData object containing the density of states.
        """
        from ase.spectrum.dosdata import RawDOSData

        # dos = self.dos(kpts, npts, delta, indices)
        kpts_kc = monkhorst_pack(kpts)
        if indices is None:
            # Return the total DOS
            omega_w = self.band_structure(kpts_kc, verbose=verbose).ravel()
            dos = RawDOSData(omega_w, np.ones_like(omega_w))
        else:
            # Return a partial DOS
            omegas, amplitudes = self.band_structure(kpts_kc,
                                                     modes=True,
                                                     verbose=verbose)
            # omegas.shape = (k-points, bands)
            # amplitudes.shape = (k-points, bands, atoms, 3)
            ampl_sq = (np.abs(amplitudes)**2).sum(axis=3)
            assert ampl_sq.ndim == 3
            assert ampl_sq.shape == omegas.shape + (len(self.indices),)
            weights = ampl_sq[:, :, indices].sum(axis=2) / ampl_sq.sum(axis=2)
            dos = RawDOSData(omegas.ravel(), weights.ravel())
        return dos

    @deprecated('Please use Phonons.get_dos() instead of Phonons.dos().')
    def dos(self, kpts=(10, 10, 10), npts=1000, delta=1e-3):
        """Calculate phonon dos as a function of energy.

        Parameters:

        kpts: tuple
            Shape of Monkhorst-Pack grid for sampling the Brillouin zone.
        npts: int
            Number of energy points.
        delta: float
            Broadening of Lorentzian line-shape in eV.

        Returns:
            Tuple of (frequencies, dos).  The frequencies are in units of eV.

        .. deprecated:: 3.23.1
            Please use the ``.get_dos()`` method instead, it returns a proper
            RawDOSData object.
        """

        # Monkhorst-Pack grid
        kpts_kc = monkhorst_pack(kpts)
        N = np.prod(kpts)
        # Get frequencies
        omega_kl = self.band_structure(kpts_kc)
        # Energy axis and dos
        omega_e = np.linspace(0., np.amax(omega_kl) + 5e-3, num=npts)
        dos_e = np.zeros_like(omega_e)

        # Sum up contribution from all q-points and branches
        for omega_l in omega_kl:
            diff_el = (omega_e[:, np.newaxis] - omega_l[np.newaxis, :])**2
            dos_el = 1. / (diff_el + (0.5 * delta)**2)
            dos_e += dos_el.sum(axis=1)

        dos_e *= 1. / (N * pi) * 0.5 * delta

        return omega_e, dos_e

    def write_modes(self, q_c, branches=0, kT=units.kB * 300, born=False,
                    repeat=(1, 1, 1), nimages=30, center=False):
        """Write modes to trajectory file.

        Parameters:

        q_c: ndarray of shape (3,)
            q-vector of the modes.
        branches: int or list
            Branch index of modes.
        kT: float
            Temperature in units of eV. Determines the amplitude of the atomic
            displacements in the modes.
        born: bool
            Include non-analytic contribution to the force constants at q -> 0.
        repeat: tuple
            Repeat atoms (l, m, n) times in the directions of the lattice
            vectors. Displacements of atoms in repeated cells carry a Bloch
            phase factor given by the q-vector and the cell lattice vector R_m.
        nimages: int
            Number of images in an oscillation.
        center: bool
            Center atoms in unit cell if True (default: False).

        To exaggerate the amplitudes for better visualization, multiply
        kT by the square of the desired factor.
        """

        if isinstance(branches, int):
            branch_l = [branches]
        else:
            branch_l = list(branches)

        # Calculate modes
        omega_l, u_l = self.band_structure([q_c], modes=True, born=born)
        # Repeat atoms
        atoms = self.atoms * repeat
        # Center
        if center:
            atoms.center()

        # Here ``Na`` refers to a composite unit cell/atom dimension
        pos_Nav = atoms.get_positions()
        # Total number of unit cells
        N = np.prod(repeat)

        # Corresponding lattice vectors R_m
        R_cN = np.indices(repeat).reshape(3, -1)
        # Bloch phase
        phase_N = np.exp(2.j * pi * np.dot(q_c, R_cN))
        phase_Na = phase_N.repeat(len(self.atoms))

        hbar = units._hbar * units.J * units.second
        for lval in branch_l:

            omega = omega_l[0, lval]
            u_av = u_l[0, lval]
            assert u_av.ndim == 2

            # For a classical harmonic oscillator, <x^2> = k T / m omega^2
            # and <x^2> = 1/2 u^2 where u is the amplitude and m is the
            # effective mass of the mode.
            # The reciprocal mass is already included in the normalization
            # of the modes.  The variable omega is actually hbar*omega (it
            # is in eV, not reciprocal ASE time units).
            u_av *= hbar * sqrt(2 * kT) / abs(omega)

            mode_av = np.zeros((len(self.atoms), 3), dtype=complex)
            # Insert slice with atomic displacements for the included atoms
            mode_av[self.indices] = u_av
            # Repeat and multiply by Bloch phase factor
            mode_Nav = np.vstack(N * [mode_av]) * phase_Na[:, np.newaxis]

            with Trajectory('%s.mode.%d.traj'
                            % (self.name, lval), 'w') as traj:
                for x in np.linspace(0, 2 * pi, nimages, endpoint=False):
                    atoms.set_positions((pos_Nav + np.exp(1.j * x) *
                                         mode_Nav).real)
                    traj.write(atoms)



import numpy as np

import ase


def test_mic():

    tol = 1e-9
    cell = np.array([[1., 0., 0.],
                     [0.5, np.sqrt(3) / 2, 0.],
                     [0., 0., 1.]]) * 10

    pos = np.dot(np.array([[0.0, 0.0, 0.0],
                           [0.5, 0.5, 0.5],
                           [0.2, 0.2, 0.2],
                           [0.25, 0.5, 0.0]]), cell)

    a = ase.Atoms('C4', pos, cell=cell, pbc=True)

    rpos = a.get_scaled_positions()

    # non-mic distance between atom 0 and 1
    d01F = np.linalg.norm(np.dot(rpos[1], cell))
    # mic distance between atom 0 (image [0,1,0]) and 1
    d01T = np.linalg.norm(np.dot(rpos[1] - np.array([0, 1, 0]), cell))
    d02F = np.linalg.norm(np.dot(rpos[2], cell))
    d02T = d02F
    # non-mic distance between atom 0 and 3
    d03F = np.linalg.norm(np.dot(rpos[3], cell))
    # mic distance between atom 0 (image [0,1,0]) and 3
    d03T = np.linalg.norm(np.dot(rpos[3] - np.array([0, 1, 0]), cell))

    # get_distance(mic=False)
    assert abs(a.get_distance(0, 1, mic=False) - d01F) < tol
    assert abs(a.get_distance(0, 2, mic=False) - d02F) < tol
    assert abs(a.get_distance(0, 3, mic=False) - d03F) < tol

    # get_distance(mic=True)
    assert abs(a.get_distance(0, 1, mic=True) - d01T) < tol
    assert abs(a.get_distance(0, 2, mic=True) - d02T) < tol
    assert abs(a.get_distance(0, 3, mic=True) - d03T) < tol

    # get_distance(mic=False, vector=True)
    assert all(abs(a.get_distance(0, 1, mic=False, vector=True)
                   - np.array([7.5, np.sqrt(18.75), 5.0])) < tol)
    assert all(abs(a.get_distance(0, 2, mic=False, vector=True)
                   - np.array([3., np.sqrt(3.), 2.0])) < tol)

    # get_distance(mic=True, vector=True)
    assert np.all(abs(a.get_distance(0, 1, mic=True, vector=True)
                      - np.array([-2.5, np.sqrt(18.75), -5.0])) < tol)
    assert np.all(abs(a.get_distance(0, 2, mic=True, vector=True)
                      - np.array([3., np.sqrt(3.), 2.0])) < tol)

    # get_all_distances(mic=False)
    all_dist = a.get_all_distances(mic=False)
    assert abs(all_dist[0, 1] - d01F) < tol
    assert abs(all_dist[0, 2] - d02F) < tol
    assert abs(all_dist[0, 3] - d03F) < tol
    assert all(abs(np.diagonal(all_dist)) < tol)

    # get_all_distances(mic=True)
    all_dist_mic = a.get_all_distances(mic=True)
    assert abs(all_dist_mic[0, 1] - d01T) < tol
    assert abs(all_dist_mic[0, 2] - d02T) < tol
    assert abs(all_dist_mic[0, 3] - d03T) < tol
    assert all(abs(np.diagonal(all_dist)) < tol)

    # get_distances(mic=False)
    for i in range(4):
        assert all(abs(a.get_distances(i, [0, 1, 2, 3], mic=False) -
                       all_dist[i]) < tol)

    # get_distances(mic=True)
    assert all(abs(a.get_distances(0, [0, 1, 2, 3], mic=True)
                   - all_dist_mic[0]) < tol)
    assert all(abs(a.get_distances(1, [0, 1, 2, 3], mic=True)
                   - all_dist_mic[1]) < tol)
    assert all(abs(a.get_distances(2, [0, 1, 2, 3], mic=True)
                   - all_dist_mic[2]) < tol)
    assert all(abs(a.get_distances(3, [0, 1, 2, 3], mic=True)
                   - all_dist_mic[3]) < tol)

    # get_distances(mic=False, vec=True)
    assert np.all(abs(a.get_distances(0, [0, 1, 2, 3], mic=False, vector=True)
                      - np.array([a.get_distance(0, i, vector=True)
                                  for i in [0, 1, 2, 3]])) < tol)
    assert np.all(abs(a.get_distances(1, [0, 1, 2, 3], mic=False, vector=True)
                      - np.array([a.get_distance(1, i, vector=True)
                                  for i in [0, 1, 2, 3]])) < tol)
    assert np.all(abs(a.get_distances(2, [0, 1, 2, 3], mic=False, vector=True)
                      - np.array([a.get_distance(2, i, vector=True)
                                  for i in [0, 1, 2, 3]])) < tol)
    assert np.all(abs(a.get_distances(3, [0, 1, 2, 3], mic=False, vector=True)
                      - np.array([a.get_distance(3, i, vector=True)
                                  for i in [0, 1, 2, 3]])) < tol)

    # get_distances(mic=True, vec=True)
    assert np.all(abs(a.get_distances(0, [0, 1, 2, 3], mic=True, vector=True)
                      - np.array([a.get_distance(0, i, mic=True, vector=True)
                                  for i in [0, 1, 2, 3]])) < tol)
    assert np.all(abs(a.get_distances(1, [0, 1, 2, 3], mic=True, vector=True)
                      - np.array([a.get_distance(1, i, mic=True, vector=True)
                                  for i in [0, 1, 2, 3]])) < tol)
    assert np.all(abs(a.get_distances(2, [0, 1, 2, 3], mic=True, vector=True)
                      - np.array([a.get_distance(2, i, mic=True, vector=True)
                                  for i in [0, 1, 2, 3]])) < tol)
    assert np.all(abs(a.get_distances(3, [0, 1, 2, 3], mic=True, vector=True)
                      - np.array([a.get_distance(3, i, mic=True, vector=True)
                                  for i in [0, 1, 2, 3]])) < tol)

    # set_distance
    a.set_distance(0, 1, 11., mic=False)
    assert abs(a.get_distance(0, 1, mic=False) - 11.) < tol
    assert abs(a.get_distance(0, 1, mic=True) - np.sqrt(46)) < tol

    # set_distance(mic=True)
    a.set_distance(0, 1, 3., mic=True)
    assert abs(a.get_distance(0, 1, mic=True) - 3.) < tol



from ase import Atoms
from ase.calculators.emt import EMT
from ase.calculators.fd import calculate_numerical_forces


def test_h2():

    h2 = Atoms('H2', positions=[(0, 0, 0), (0, 0, 1.1)],
               calculator=EMT())
    f1 = calculate_numerical_forces(h2, 0.0001)
    f2 = h2.get_forces()
    assert abs(f1 - f2).max() < 1e-6



#!/usr/bin/env python3
"""Test Periodic Table."""
from ase.utils.ptable import ptable

spacing = 1.0


def test_pourbaix():
    """Test the periodic table generating function."""
    atoms = ptable(spacing)
    x = atoms.positions[:, 0]
    # testing to ensure that the table is 18 groups wide
    assert spacing * 17 == (x.max() - x.min())
    # testing that there are 118 unique elements
    symbols = set(atoms.get_chemical_symbols())
    assert len(symbols) == 118



import numpy as np
import pytest
from numpy.testing import assert_allclose

import ase.lattice
from ase.spacegroup import (
    Spacegroup,
    get_bravais_class,
    get_point_group,
    polar_space_group,
)
from ase.spacegroup.spacegroup import SpacegroupNotFoundError

TOL = 1E-10

functions = [get_bravais_class, get_point_group, polar_space_group]


@pytest.mark.parametrize("sg,lattice,point_group,polar",
                         [[100, ase.lattice.TET, '4mm', True],
                          [225, ase.lattice.FCC, '4/m -3 2/m', False]])
def test_valid_spacegroup(sg, lattice, point_group, polar):
    assert get_bravais_class(sg) == lattice
    assert get_point_group(sg) == point_group
    assert polar_space_group(sg) == polar


@pytest.mark.parametrize("func", functions)
def test_nonpositive_spacegroup(func):
    with pytest.raises(ValueError, match="positive"):
        func(0)


@pytest.mark.parametrize("func", functions)
def test_bad_spacegroup(func):
    with pytest.raises(ValueError, match="Bad"):
        func(400)


def _spacegroup_reciprocal_cell(no, setting):
    try:
        sg = Spacegroup(no, setting)
    except SpacegroupNotFoundError:
        return
    reciprocal_check = np.linalg.inv(sg.scaled_primitive_cell).T
    assert_allclose(sg.reciprocal_cell, reciprocal_check, atol=TOL)


def test_spacegroup_reciprocal_cell():
    # Would be better to use pytest.mark.parametrize if we can figure out
    # how to not list hundreds of test cases in the output.
    for setting in [1, 2]:
        for no in range(1, 231):
            _spacegroup_reciprocal_cell(setting, no)



import numpy as np

from ase.dft import get_distribution_moment


def test_distmom():

    precision = 1E-8

    x = np.linspace(-50., 50., 1000)
    y = np.exp(-x**2 / 2.)
    area, center, mom2 = get_distribution_moment(x, y, (0, 1, 2))
    assert sum((abs(area - np.sqrt(2. * np.pi)),
               abs(center), abs(mom2 - 1.))) < precision

    x = np.linspace(-1., 1., 100000)
    for order in range(9):
        y = x**order
        area = get_distribution_moment(x, y)
        assert abs(area - (1. - (-1.)**(order + 1)) / (order + 1.)) < precision

    x = np.linspace(-50., 50., 100)
    y = np.exp(-2. * (x - 7.)**2 / 10.) + np.exp(-2. * (x + 5.)**2 / 10.)
    center = get_distribution_moment(x, y, 1)
    assert abs(center - 1.) < precision



import pytest

from ase import Atoms
from ase.build.supercells import make_supercell
from ase.calculators.emt import EMT
from ase.lattice import all_variants


def emt_energy_per_atom(atoms):
    atoms = atoms.copy()
    atoms.calc = EMT()
    return atoms.get_potential_energy() / len(atoms)


@pytest.mark.parametrize('lat', [var for var in all_variants()
                                 if var.ndim == 3])
def test_conventional_map(lat):
    if not hasattr(lat, 'conventional_cellmap'):
        pytest.skip()

    conv_lat = lat.conventional()
    prim_atoms = Atoms('Au', cell=lat.tocell(), pbc=1)
    conv_atoms = make_supercell(prim_atoms, lat.conventional_cellmap)

    e1 = emt_energy_per_atom(prim_atoms)
    e2 = emt_energy_per_atom(conv_atoms)

    assert e1 == pytest.approx(e2)
    assert conv_lat.cellpar() == pytest.approx(conv_atoms.cell.cellpar())

    # Rule out also that cells could differ by a rotation:
    assert conv_lat.tocell()[:] == pytest.approx(conv_atoms.cell[:])



# Note:
# Try to avoid module level import statements here to reduce
# import time during CLI execution


class CLICommand:
    """ Execute code on files.

    The given python code is evaluated on the Atoms object read from
    the input file for each frame of the file. Either of -e or -E
    option should provided for evaluating code given as a string or
    from a file, respectively.

    Variables which can be used inside the python code:
    - `index`: Index of the current Atoms object.
    - `atoms`: Current Atoms object.
    - `images`: List of all images given as input.
    """

    @staticmethod
    def add_arguments(parser):
        add = parser.add_argument
        add('input', nargs='+', metavar='input-file')
        add('-e', '--exec-code',
            help='Python code to execute on each atoms. The Atoms'
            ' object is available as `atoms`. '
            'Example: For printing cell parameters from all the '
            'frames, `print(atoms.cell.cellpar())`')
        add('-E', '--exec-file',
            help='Python source code file to execute on each '
            'frame, usage is as for -e/--exec-code.')
        add('-i', '--input-format', metavar='FORMAT',
            help='Specify input FORMAT')
        add('-n', '--image-number',
            default=':', metavar='NUMBER',
            help='Pick images from trajectory.  NUMBER can be a '
            'single number (use a negative number to count from '
            'the back) or a range: start:stop:step, where the '
            '":step" part can be left out - default values are '
            '0:nimages:1.')
        add('--read-args', nargs='+', action='store',
            default={}, metavar="KEY=VALUE",
            help='Additional keyword arguments to pass to '
            '`ase.io.read()`.')

    @staticmethod
    def run(args, parser):
        import runpy

        from ase.io import read

        if not (args.exec_code or args.exec_file):
            parser.error("At least one of '-e' or '-E' must be provided")

        if args.read_args:
            args.read_args = eval("dict({})"
                                  .format(', '.join(args.read_args)))

        configs = []
        for filename in args.input:
            atoms = read(filename, args.image_number,
                         format=args.input_format, **args.read_args)
            if isinstance(atoms, list):
                configs.extend(atoms)
            else:
                configs.append(atoms)

        variables = {'images': configs}
        for index, atoms in enumerate(configs):
            variables['atoms'] = atoms
            variables['index'] = index
            if args.exec_code:
                exec(compile(args.exec_code, '<string>', 'exec'), variables)
            if args.exec_file:
                runpy.run_path(args.exec_file, init_globals=variables)



import errno
import functools
import io
import os
import pickle
import re
import string
import sys
import time
import warnings
from contextlib import ExitStack, contextmanager
from importlib import import_module
from math import atan2, cos, degrees, gcd, radians, sin
from pathlib import Path, PurePath
from typing import Callable, Dict, List, Type, Union

import numpy as np

from ase.formula import formula_hill, formula_metal

__all__ = ['basestring', 'import_module', 'seterr', 'plural',
           'devnull', 'gcd', 'convert_string_to_fd', 'Lock',
           'opencew', 'OpenLock', 'rotate', 'irotate', 'pbc2pbc', 'givens',
           'hsv2rgb', 'hsv', 'pickleload', 'reader',
           'formula_hill', 'formula_metal', 'PurePath', 'xwopen',
           'tokenize_version', 'get_python_package_path_description']


def tokenize_version(version_string: str):
    """Parse version string into a tuple for version comparisons.

    Usage: tokenize_version('3.8') < tokenize_version('3.8.1').
    """
    tokens = []
    for component in version_string.split('.'):
        match = re.match(r'(\d*)(.*)', component)
        assert match is not None, f'Cannot parse component {component}'
        number_str, tail = match.group(1, 2)
        try:
            number = int(number_str)
        except ValueError:
            number = -1
        tokens += [number, tail]
    return tuple(tokens)


# Python 2+3 compatibility stuff (let's try to remove these things):
basestring = str
pickleload = functools.partial(pickle.load, encoding='bytes')


def deprecated(
    message: Union[str, Warning],
    category: Type[Warning] = FutureWarning,
    callback: Callable[[List, Dict], bool] = lambda args, kwargs: True
):
    """Return a decorator deprecating a function.

    Parameters
    ----------
    message : str or Warning
        The message to be emitted. If ``message`` is a Warning, then
        ``category`` is ignored and ``message.__class__`` will be used.
    category : Type[Warning], default=FutureWarning
        The type of warning to be emitted. If ``message`` is a ``Warning``
        instance, then ``category`` will be ignored and ``message.__class__``
        will be used.
    callback : Callable[[List, Dict], bool], default=lambda args, kwargs: True
        A callable that determines if the warning should be emitted and handles
        any processing prior to calling the deprecated function. The callable
        will receive two arguments, a list and a dictionary. The list will
        contain the positional arguments that the deprecated function was
        called with at runtime while the dictionary will contain the keyword
        arguments. The callable *must* return ``True`` if the warning is to be
        emitted and ``False`` otherwise. The list and dictionary will be
        unpacked into the positional and keyword arguments, respectively, used
        to call the deprecated function.

    Returns
    -------
    deprecated_decorator : Callable
        A decorator for deprecated functions that can be used to conditionally
        emit deprecation warnings and/or pre-process the arguments of a
        deprecated function.

    Example
    -------
    >>> # Inspect & replace a keyword parameter passed to a deprecated function
    >>> from typing import Any, Callable, Dict, List
    >>> import warnings
    >>> from ase.utils import deprecated

    >>> def alias_callback_factory(kwarg: str, alias: str) -> Callable:
    ...     def _replace_arg(_: List, kwargs: Dict[str, Any]) -> bool:
    ...         kwargs[kwarg] = kwargs[alias]
    ...         del kwargs[alias]
    ...         return True
    ...     return _replace_arg

    >>> MESSAGE = ("Calling this function with `atoms` is deprecated. "
    ...            "Use `optimizable` instead.")
    >>> @deprecated(
    ...     MESSAGE,
    ...     category=DeprecationWarning,
    ...     callback=alias_callback_factory("optimizable", "atoms")
    ... )
    ... def function(*, atoms=None, optimizable=None):
    ...     '''
    ...     .. deprecated:: 3.23.0
    ...         Calling this function with ``atoms`` is deprecated.
    ...         Use ``optimizable`` instead.
    ...     '''
    ...     print(f"atoms: {atoms}")
    ...     print(f"optimizable: {optimizable}")

    >>> with warnings.catch_warnings(record=True) as w:
    ...     warnings.simplefilter("always")
    ...     function(atoms="atoms")
    atoms: None
    optimizable: atoms

    >>> w[-1].category == DeprecationWarning
    True
    """

    def deprecated_decorator(func):
        @functools.wraps(func)
        def deprecated_function(*args, **kwargs):
            _args = list(args)
            if callback(_args, kwargs):
                warnings.warn(message, category=category, stacklevel=2)

            return func(*_args, **kwargs)

        return deprecated_function

    return deprecated_decorator


@contextmanager
def seterr(**kwargs):
    """Set how floating-point errors are handled.

    See np.seterr() for more details.
    """
    old = np.seterr(**kwargs)
    try:
        yield
    finally:
        np.seterr(**old)


def plural(n, word):
    """Use plural for n!=1.

    >>> from ase.utils import plural

    >>> plural(0, 'egg'), plural(1, 'egg'), plural(2, 'egg')
    ('0 eggs', '1 egg', '2 eggs')
    """
    if n == 1:
        return '1 ' + word
    return '%d %ss' % (n, word)


class DevNull:
    encoding = 'UTF-8'
    closed = False

    _use_os_devnull = deprecated('use open(os.devnull) instead',
                                 DeprecationWarning)
    # Deprecated for ase-3.21.0.  Change to futurewarning later on.

    @_use_os_devnull
    def write(self, string):
        pass

    @_use_os_devnull
    def flush(self):
        pass

    @_use_os_devnull
    def seek(self, offset, whence=0):
        return 0

    @_use_os_devnull
    def tell(self):
        return 0

    @_use_os_devnull
    def close(self):
        pass

    @_use_os_devnull
    def isatty(self):
        return False

    @_use_os_devnull
    def read(self, n=-1):
        return ''


devnull = DevNull()


@deprecated('convert_string_to_fd does not facilitate proper resource '
            'management.  '
            'Please use e.g. ase.utils.IOContext class instead.')
def convert_string_to_fd(name, world=None):
    """Create a file-descriptor for text output.

    Will open a file for writing with given name.  Use None for no output and
    '-' for sys.stdout.

    .. deprecated:: 3.22.1
        Please use e.g. :class:`ase.utils.IOContext` class instead.
    """
    if world is None:
        from ase.parallel import world
    if name is None or world.rank != 0:
        return open(os.devnull, 'w')
    if name == '-':
        return sys.stdout
    if isinstance(name, (str, PurePath)):
        return open(str(name), 'w')  # str for py3.5 pathlib
    return name  # we assume name is already a file-descriptor


# Only Windows has O_BINARY:
CEW_FLAGS = os.O_CREAT | os.O_EXCL | os.O_WRONLY | getattr(os, 'O_BINARY', 0)


@contextmanager
def xwopen(filename, world=None):
    """Create and open filename exclusively for writing.

    If master cpu gets exclusive write access to filename, a file
    descriptor is returned (a dummy file descriptor is returned on the
    slaves).  If the master cpu does not get write access, None is
    returned on all processors."""

    fd = opencew(filename, world)
    try:
        yield fd
    finally:
        if fd is not None:
            fd.close()


# @deprecated('use "with xwopen(...) as fd: ..." to prevent resource leak')
def opencew(filename, world=None):
    return _opencew(filename, world)


def _opencew(filename, world=None):
    import ase.parallel as parallel
    if world is None:
        world = parallel.world

    closelater = []

    def opener(file, flags):
        return os.open(file, flags | CEW_FLAGS)

    try:
        error = 0
        if world.rank == 0:
            try:
                fd = open(filename, 'wb', opener=opener)
            except OSError as ex:
                error = ex.errno
            else:
                closelater.append(fd)
        else:
            fd = open(os.devnull, 'wb')
            closelater.append(fd)

        # Synchronize:
        error = world.sum_scalar(error)
        if error == errno.EEXIST:
            return None
        if error:
            raise OSError(error, 'Error', filename)

        return fd
    except BaseException:
        for fd in closelater:
            fd.close()
        raise


def opencew_text(*args, **kwargs):
    fd = opencew(*args, **kwargs)
    if fd is None:
        return None
    return io.TextIOWrapper(fd)


class Lock:
    def __init__(self, name='lock', world=None, timeout=float('inf')):
        self.name = str(name)
        self.timeout = timeout
        if world is None:
            from ase.parallel import world
        self.world = world

    def acquire(self):
        dt = 0.2
        t1 = time.time()
        while True:
            fd = opencew(self.name, self.world)
            if fd is not None:
                self.fd = fd
                break
            time_left = self.timeout - (time.time() - t1)
            if time_left <= 0:
                raise TimeoutError
            time.sleep(min(dt, time_left))
            dt *= 2

    def release(self):
        self.world.barrier()
        # Important to close fd before deleting file on windows
        # as a WinError would otherwise be raised.
        self.fd.close()
        if self.world.rank == 0:
            os.remove(self.name)
        self.world.barrier()

    def __enter__(self):
        self.acquire()

    def __exit__(self, type, value, tb):
        self.release()


class OpenLock:
    def acquire(self):
        pass

    def release(self):
        pass

    def __enter__(self):
        pass

    def __exit__(self, type, value, tb):
        pass


def search_current_git_hash(arg, world=None):
    """Search for .git directory and current git commit hash.

    Parameters:

    arg: str (directory path) or python module
        .git directory is searched from the parent directory of
        the given directory or module.
    """
    if world is None:
        from ase.parallel import world
    if world.rank != 0:
        return None

    # Check argument
    if isinstance(arg, str):
        # Directory path
        dpath = arg
    else:
        # Assume arg is module
        dpath = os.path.dirname(arg.__file__)
    # dpath = os.path.abspath(dpath)
    # in case this is just symlinked into $PYTHONPATH
    dpath = os.path.realpath(dpath)
    dpath = os.path.dirname(dpath)  # Go to the parent directory
    git_dpath = os.path.join(dpath, '.git')
    if not os.path.isdir(git_dpath):
        # Replace this 'if' with a loop if you want to check
        # further parent directories
        return None
    HEAD_file = os.path.join(git_dpath, 'HEAD')
    if not os.path.isfile(HEAD_file):
        return None
    with open(HEAD_file) as fd:
        line = fd.readline().strip()
    if line.startswith('ref: '):
        ref = line[5:]
        ref_file = os.path.join(git_dpath, ref)
    else:
        # Assuming detached HEAD state
        ref_file = HEAD_file
    if not os.path.isfile(ref_file):
        return None
    with open(ref_file) as fd:
        line = fd.readline().strip()
    if all(c in string.hexdigits for c in line):
        return line
    return None


def rotate(rotations, rotation=np.identity(3)):
    """Convert string of format '50x,-10y,120z' to a rotation matrix.

    Note that the order of rotation matters, i.e. '50x,40z' is different
    from '40z,50x'.
    """

    if rotations == '':
        return rotation.copy()

    for i, a in [('xyz'.index(s[-1]), radians(float(s[:-1])))
                 for s in rotations.split(',')]:
        s = sin(a)
        c = cos(a)
        if i == 0:
            rotation = np.dot(rotation, [(1, 0, 0),
                                         (0, c, s),
                                         (0, -s, c)])
        elif i == 1:
            rotation = np.dot(rotation, [(c, 0, -s),
                                         (0, 1, 0),
                                         (s, 0, c)])
        else:
            rotation = np.dot(rotation, [(c, s, 0),
                                         (-s, c, 0),
                                         (0, 0, 1)])
    return rotation


def givens(a, b):
    """Solve the equation system::

      [ c s]   [a]   [r]
      [    ] . [ ] = [ ]
      [-s c]   [b]   [0]
    """
    sgn = np.sign
    if b == 0:
        c = sgn(a)
        s = 0
        r = abs(a)
    elif abs(b) >= abs(a):
        cot = a / b
        u = sgn(b) * (1 + cot**2)**0.5
        s = 1. / u
        c = s * cot
        r = b * u
    else:
        tan = b / a
        u = sgn(a) * (1 + tan**2)**0.5
        c = 1. / u
        s = c * tan
        r = a * u
    return c, s, r


def irotate(rotation, initial=np.identity(3)):
    """Determine x, y, z rotation angles from rotation matrix."""
    a = np.dot(initial, rotation)
    cx, sx, rx = givens(a[2, 2], a[1, 2])
    cy, sy, _ry = givens(rx, a[0, 2])
    cz, sz, _rz = givens(cx * a[1, 1] - sx * a[2, 1],
                        cy * a[0, 1] - sy * (sx * a[1, 1] + cx * a[2, 1]))
    x = degrees(atan2(sx, cx))
    y = degrees(atan2(-sy, cy))
    z = degrees(atan2(sz, cz))
    return x, y, z


def pbc2pbc(pbc):
    newpbc = np.empty(3, bool)
    newpbc[:] = pbc
    return newpbc


def string2index(stridx: str) -> Union[int, slice, str]:
    """Convert index string to either int or slice"""
    if ':' not in stridx:
        # may contain database accessor
        try:
            return int(stridx)
        except ValueError:
            return stridx
    i = [None if s == '' else int(s) for s in stridx.split(':')]
    return slice(*i)


def hsv2rgb(h, s, v):
    """http://en.wikipedia.org/wiki/HSL_and_HSV

    h (hue) in [0, 360[
    s (saturation) in [0, 1]
    v (value) in [0, 1]

    return rgb in range [0, 1]
    """
    if v == 0:
        return 0, 0, 0
    if s == 0:
        return v, v, v

    i, f = divmod(h / 60., 1)
    p = v * (1 - s)
    q = v * (1 - s * f)
    t = v * (1 - s * (1 - f))

    if i == 0:
        return v, t, p
    elif i == 1:
        return q, v, p
    elif i == 2:
        return p, v, t
    elif i == 3:
        return p, q, v
    elif i == 4:
        return t, p, v
    elif i == 5:
        return v, p, q
    else:
        raise RuntimeError('h must be in [0, 360]')


def hsv(array, s=.9, v=.9):
    array = (array + array.min()) * 359. / (array.max() - array.min())
    result = np.empty((len(array.flat), 3))
    for rgb, h in zip(result, array.flat):
        rgb[:] = hsv2rgb(h, s, v)
    return np.reshape(result, array.shape + (3,))


# This code does the same, but requires pylab
# def cmap(array, name='hsv'):
#     import pylab
#     a = (array + array.min()) / array.ptp()
#     rgba = getattr(pylab.cm, name)(a)
#     return rgba[:-1] # return rgb only (not alpha)


def longsum(x):
    """128-bit floating point sum."""
    return float(np.asarray(x, dtype=np.longdouble).sum())


@contextmanager
def workdir(path, mkdir=False):
    """Temporarily change, and optionally create, working directory."""
    path = Path(path)
    if mkdir:
        path.mkdir(parents=True, exist_ok=True)

    olddir = os.getcwd()
    os.chdir(path)
    try:
        yield  # Yield the Path or dirname maybe?
    finally:
        os.chdir(olddir)


class iofunction:
    """Decorate func so it accepts either str or file.

    (Won't work on functions that return a generator.)"""

    def __init__(self, mode):
        self.mode = mode

    def __call__(self, func):
        @functools.wraps(func)
        def iofunc(file, *args, **kwargs):
            openandclose = isinstance(file, (str, PurePath))
            fd = None
            try:
                if openandclose:
                    fd = open(str(file), self.mode)
                else:
                    fd = file
                obj = func(fd, *args, **kwargs)
                return obj
            finally:
                if openandclose and fd is not None:
                    # fd may be None if open() failed
                    fd.close()
        return iofunc


def writer(func):
    return iofunction('w')(func)


def reader(func):
    return iofunction('r')(func)


# The next two functions are for hotplugging into a JSONable class
# using the jsonable decorator.  We are supposed to have this kind of stuff
# in ase.io.jsonio, but we'd rather import them from a 'basic' module
# like ase/utils than one which triggers a lot of extra (cyclic) imports.

def write_json(self, fd):
    """Write to JSON file."""
    from ase.io.jsonio import write_json as _write_json
    _write_json(fd, self)


@classmethod  # type: ignore[misc]
def read_json(cls, fd):
    """Read new instance from JSON file."""
    from ase.io.jsonio import read_json as _read_json
    obj = _read_json(fd)
    assert isinstance(obj, cls)
    return obj


def jsonable(name):
    """Decorator for facilitating JSON I/O with a class.

    Pokes JSON-based read and write functions into the class.

    In order to write an object to JSON, it needs to be a known simple type
    (such as ndarray, float, ...) or implement todict().  If the class
    defines a string called ase_objtype, the decoder will want to convert
    the object back into its original type when reading."""
    def jsonableclass(cls):
        cls.ase_objtype = name
        if not hasattr(cls, 'todict'):
            raise TypeError('Class must implement todict()')

        # We may want the write and read to be optional.
        # E.g. a calculator might want to be JSONable, but not
        # that .write() produces a JSON file.
        #
        # This is mostly for 'lightweight' object IO.
        cls.write = write_json
        cls.read = read_json
        return cls
    return jsonableclass


class ExperimentalFeatureWarning(Warning):
    pass


def experimental(func):
    """Decorator for functions not ready for production use."""
    @functools.wraps(func)
    def expfunc(*args, **kwargs):
        warnings.warn('This function may change or misbehave: {}()'
                      .format(func.__qualname__),
                      ExperimentalFeatureWarning)
        return func(*args, **kwargs)
    return expfunc


def lazymethod(meth):
    """Decorator for lazy evaluation and caching of data.

    Example::

      class MyClass:

         @lazymethod
         def thing(self):
             return expensive_calculation()

    The method body is only executed first time thing() is called, and
    its return value is stored.  Subsequent calls return the cached
    value."""
    name = meth.__name__

    @functools.wraps(meth)
    def getter(self):
        try:
            cache = self._lazy_cache
        except AttributeError:
            cache = self._lazy_cache = {}

        if name not in cache:
            cache[name] = meth(self)
        return cache[name]
    return getter


def atoms_to_spglib_cell(atoms):
    """Convert atoms into data suitable for calling spglib."""
    return (atoms.get_cell(),
            atoms.get_scaled_positions(),
            atoms.get_atomic_numbers())


def warn_legacy(feature_name):
    warnings.warn(
        f'The {feature_name} feature is untested and ASE developers do not '
        'know whether it works or how to use it.  Please rehabilitate it '
        '(by writing unittests) or it may be removed.',
        FutureWarning)


def lazyproperty(meth):
    """Decorator like lazymethod, but making item available as a property."""
    return property(lazymethod(meth))


class IOContext:
    @lazyproperty
    def _exitstack(self):
        return ExitStack()

    def __enter__(self):
        return self

    def __exit__(self, *args):
        self.close()

    def closelater(self, fd):
        return self._exitstack.enter_context(fd)

    def close(self):
        self._exitstack.close()

    def openfile(self, file, comm, mode='w'):
        if hasattr(file, 'close'):
            return file  # File already opened, not for us to close.

        encoding = None if mode.endswith('b') else 'utf-8'

        if file is None or comm.rank != 0:
            return self.closelater(open(os.devnull, mode=mode,
                                        encoding=encoding))

        if file == '-':
            return sys.stdout

        return self.closelater(open(file, mode=mode, encoding=encoding))


def get_python_package_path_description(
        package, default='module has no path') -> str:
    """Helper to get path description of a python package/module

    If path has multiple elements, the first one is returned.
    If it is empty, the default is returned.
    Exceptions are returned as strings default+(exception).
    Always returns a string.
    """
    try:
        p = list(package.__path__)
        if p:
            return str(p[0])
        else:
            return default
    except Exception as ex:
        return f"{default} ({ex})"
